import torch
import torch.nn as nn
from itertools import islice
import numpy as np
import random
from vit_prisma.training.training_dictionary import loss_function_dict
from abc import ABC, abstractmethod

class PrismaCallback(ABC):
    @abstractmethod
    def on_epoch_end(self, epoch, net, val_loader, wandb_logger):
        pass

    @abstractmethod
    def on_step_end(self, step, net, val_loader, wandb_logger):
        pass

def calculate_accuracy(net, data_loader, device, N=2000, batch_size=50):
    net.eval()
    with torch.no_grad():
        correct = 0
        total = 0
        for items in islice(data_loader, N // batch_size):
            x, labels, *extras = items
            logits = net(x.to(device))
            predictions = torch.argmax(logits, dim=1)
            correct += torch.sum(predictions == labels.to(device)).item()
            total += len(labels)
        return correct / total

def calculate_loss(net, data_loader, loss_fn, device, N=2000, batch_size=50):
    net.eval()
    with torch.no_grad():
        total = 0
        points = 0
        for items in islice(data_loader, N // batch_size):
            x, labels, *extras = items
            logits = net(x.to(device))
            total += loss_fn(logits, labels.to(device)).item()
            points += len(labels)
        return total / points
    
def set_seed(seed):
    torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    random.seed(seed)
    np.random.seed(seed)



class EarlyStopping:
    def __init__(self, patience=3, verbose=False, delta=0):
        self.patience = patience
        self.verbose = verbose
        self.counter = 0
        self.best_score = None
        self.early_stop = False
        self.val_accuracy_max = float('-inf')
        self.delta = delta

    def __call__(self, val_accuracy):
        score = val_accuracy

        if self.best_score is None:
            self.best_score = score
        elif score < self.best_score + self.delta:
            self.counter += 1
            print(f'EarlyStopping counter: {self.counter} out of {self.patience}')
            if self.counter >= self.patience:
                print("Early stopping triggered!")
                self.early_stop = True
        else:
            self.best_score = score
            self.counter = 0



import pytest
import torch

from vit_prisma.models.base_vit import HookedViT
from vit_prisma.configs.HookedViTConfig import HookedViTConfig

#Test taken from transformerlens with minor modifications

batch_size = 2
channels = 3
height = 224
width = 224
input_image = torch.rand((batch_size, channels, height, width))

# instantiate a untrained network
n_layers = 1
d_head = 8
d_model = 8
d_mlp=8

model = HookedViT(HookedViTConfig(n_layers,d_head,d_model, d_mlp, return_type="logits"))
embed = lambda name: name == "hook_embed"


class Counter:
    def __init__(self):
        self.count = 0

    def inc(self, *args, **kwargs):
        self.count += 1


def test_hook_attaches_normally():
    c = Counter()
    _ = model.run_with_hooks(input_image, fwd_hooks=[(embed, c.inc)])
    assert all([len(hp.fwd_hooks) == 0 for _, hp in model.hook_dict.items()])
    assert c.count == 1
    model.remove_all_hook_fns(including_permanent=True)


def test_perma_hook_attaches_normally():
    c = Counter()
    model.add_perma_hook(embed, c.inc)
    assert len(model.hook_dict["hook_embed"].fwd_hooks) == 1
    model.run_with_hooks(input_image, fwd_hooks=[])
    assert len(model.hook_dict["hook_embed"].fwd_hooks) == 1
    assert c.count == 1
    model.remove_all_hook_fns(including_permanent=True)


def test_hook_context_manager():
    c = Counter()
    with model.hooks(fwd_hooks=[(embed, c.inc)]):
        assert len(model.hook_dict["hook_embed"].fwd_hooks) == 1
        model.forward(input_image)
    assert len(model.hook_dict["hook_embed"].fwd_hooks) == 0
    assert c.count == 1
    model.remove_all_hook_fns(including_permanent=True)


def test_nested_hook_context_manager():
    c = Counter()
    with model.hooks(fwd_hooks=[(embed, c.inc)]):
        assert len(model.hook_dict["hook_embed"].fwd_hooks) == 1
        model.forward(input_image)
        assert c.count == 1
        with model.hooks(fwd_hooks=[(embed, c.inc)]):
            assert len(model.hook_dict["hook_embed"].fwd_hooks) == 2
            model.forward(input_image)
            assert c.count == 3  # 2 from outer, 1 from inner
        assert len(model.hook_dict["hook_embed"].fwd_hooks) == 1
    assert len(model.hook_dict["hook_embed"].fwd_hooks) == 0
    assert c.count == 3
    model.remove_all_hook_fns(including_permanent=True)


def test_context_manager_run_with_cache():
    c = Counter()
    with model.hooks(fwd_hooks=[(embed, c.inc)]):
        assert len(model.hook_dict["hook_embed"].fwd_hooks) == 1
        model.run_with_cache(input_image)
        assert len(model.hook_dict["hook_embed"].fwd_hooks) == 1
    assert len(model.hook_dict["hook_embed"].fwd_hooks) == 0
    assert c.count == 1
    model.remove_all_hook_fns(including_permanent=True)


def test_hook_context_manager_with_permanent_hook():
    c = Counter()
    model.add_perma_hook(embed, c.inc)
    assert len(model.hook_dict["hook_embed"].fwd_hooks) == 1
    with model.hooks(fwd_hooks=[(embed, c.inc)]):
        assert len(model.hook_dict["hook_embed"].fwd_hooks) == 2
        model.forward(input_image)
    assert len(model.hook_dict["hook_embed"].fwd_hooks) == 1
    assert c.count == 2  # 1 from permanent, 1 from context manager
    model.remove_all_hook_fns(including_permanent=True)


def test_nested_context_manager_with_failure():
    def fail_hook(z, hook):
        raise ValueError("fail")

    c = Counter()
    with model.hooks(fwd_hooks=[(embed, c.inc)]):
        with pytest.raises(ValueError):
            with model.hooks(fwd_hooks=[(embed, fail_hook)]):
                assert len(model.hook_dict["hook_embed"].fwd_hooks) == 2
                model.forward(input_image)
        assert len(model.hook_dict["hook_embed"].fwd_hooks) == 1
        assert c.count == 1
    assert len(model.hook_dict["hook_embed"].fwd_hooks) == 0
    model.remove_all_hook_fns(including_permanent=True)


def test_reset_hooks_in_context_manager():
    c = Counter()
    with model.hooks(fwd_hooks=[(embed, c.inc)]):
        assert len(model.hook_dict["hook_embed"].fwd_hooks) == 1
        model.reset_hooks()
        assert len(model.hook_dict["hook_embed"].fwd_hooks) == 0
    assert len(model.hook_dict["hook_embed"].fwd_hooks) == 0
    model.remove_all_hook_fns(including_permanent=True)


def test_remove_hook():
    c = Counter()
    model.add_perma_hook(embed, c.inc)
    assert len(model.hook_dict["hook_embed"].fwd_hooks) == 1  # 1 after adding
    model.remove_all_hook_fns()
    assert (
        len(model.hook_dict["hook_embed"].fwd_hooks) == 1
    )  # permanent not removed without flag
    model.remove_all_hook_fns(including_permanent=True)
    assert len(model.hook_dict["hook_embed"].fwd_hooks) == 0  # removed now
    model.run_with_hooks(input_image, fwd_hooks=[])
    assert c.count == 0
    model.remove_all_hook_fns(including_permanent=True)


def test_conditional_hooks():
    """Test that it's only possible to add certain hooks when certain conditions are met"""

    def identity_hook(z, hook):
        return z

    for hook_name, set_use_hook_function in [
        ("blocks.0.attn.hook_result", model.set_use_attn_result),
        ("blocks.0.hook_q_input", model.set_use_split_qkv_input),
        ("blocks.0.hook_mlp_in", model.set_use_hook_mlp_in),
        ("blocks.0.hook_attn_in", model.set_use_attn_in),
    ]:
        model.reset_hooks()
        set_use_hook_function(False)
        # Ensure that we get an error when we inappropriately add a hook
        with pytest.raises(AssertionError):
            model.add_hook(hook_name, identity_hook)
        # Ensure we DON'T get an error when we add a hook properly
        set_use_hook_function(True)
        model.add_hook(hook_name, identity_hook)

        # Reset the flag
        set_use_hook_function(False)
    # Check that hooks cache things with the right shape

    # The correct shapes of cached values for the hooks with three dimensions and with four dimensions
    # (2, 50, ... because the batch size is 2 and the sequence length is 50 (1 cls token + 7x7 patches))
    correct_shapes = {
        3: (2, 50, model.cfg.d_model),
        4: (2, 50, model.cfg.n_heads, model.cfg.d_model),
    }

    for hook_name, set_use_hook_function, number_of_dimensions in [
        ("blocks.0.hook_q_input", model.set_use_split_qkv_input, 4),
        ("blocks.0.hook_attn_in", model.set_use_attn_in, 4),
        ("blocks.0.hook_mlp_in", model.set_use_hook_mlp_in, 3),
    ]:
        model.reset_hooks()
        set_use_hook_function(True)

        cache = model.run_with_cache(
            input_image,
            names_filter=lambda x: x == hook_name,
        )[1]

        assert list(cache.keys()) == [hook_name]
        assert cache[hook_name].shape == correct_shapes[number_of_dimensions]

        # Reset the flag
        set_use_hook_function(False)


@pytest.mark.parametrize(
    "zero_attach_pos,prepend",
    [
        (zero_attach_pos, prepend)
        for zero_attach_pos in range(2)
        for prepend in [True, False]
    ],
)
def test_prepending_hooks(zero_attach_pos, prepend):
    """Add two hooks to a model: one that sets last layer activations to all 0s
    One that sets them to random noise.

    If the last activations are 0, then the logits will just be the model's logit bias.
    This is not true if the last activations are random noise.

    This test tests the prepending functionality by ensuring this property holds!"""

    def set_to_zero(z, hook):
        z[:] = 0.0
        return z

    def set_to_randn(z, hook):
        z = torch.randn_like(z) * 0.1
        return z

    model.reset_hooks()

    for hook_idx in range(2):
        model.add_hook(
            "blocks.0.hook_resid_post",
            set_to_zero if hook_idx == zero_attach_pos else set_to_randn,
            prepend=prepend,
        )
    logits = model(input_image[0][None, ...])

    logits_are_unembed_bias = (zero_attach_pos == 1) != prepend
    # the logits should be equal to the unembed bias
    # exactly when the zero hook is attached last XOR it is prepended

    assert torch.allclose(logits, model.head.b_H[None, :]) == logits_are_unembed_bias


import numpy as np
import string, random, json
import torch
import os
from typing import List
from jinja2 import Template

def convert_to_3_channels(image):
    # Check if the image has only one channel (grayscale)
    if image.shape[-1] == 1 or image.ndim == 2:
        # Stack the grayscale image three times along the third axis to make it 3-channel
        image = np.squeeze(image)
        image = np.stack([image, image, image], axis=-1)
    return image

def generate_random_string(length=10):
    '''
    Helper function to generate canvas IDs for javascript figures.
    '''
    letters = string.ascii_lowercase
    return ''.join(random.choice(letters) for i in range(length))

def prepare_image(image):
    if isinstance(image,torch.Tensor):
        image = image.numpy()
    image = (image - image.min()) / (image.max() - image.min()) * 255
    image = image.astype('uint8')
    image = np.transpose(image, (1, 2, 0))
    image = convert_to_3_channels(image)
    return image

def flatten_into_patches(image, patch_size, image_size):
    patches = [image[i:i+patch_size, j:j+patch_size, :] for i in range(0, image_size, patch_size) for j in range(0, image_size, patch_size)]
    flattened_patches = [patch.flatten().tolist() for patch in patches]
    return flattened_patches

def normalize_attn_head(attn_head):
    min_val = np.min(attn_head)
    max_val = np.max(attn_head)
    normalized_attn_head = (attn_head - min_val) / (max_val - min_val)
    return normalized_attn_head


# prep data to send to javascript
class AttentionHeadImageJSInfo:

    def __init__(self, attn_head, image, name="No Name", cls_token=True):

        normalized_ah = normalize_attn_head(attn_head)
        if not cls_token:
            normalized_ah = normalized_ah[1:, 1:]

        image_size = image.shape[-1]
        assert image_size == image.shape[-2], "images are assumed to be square"

        patch_size = int(image_size // np.sqrt(len(normalized_ah) - 1))
        image = prepare_image(image)
        flattened_patches = flatten_into_patches(image, patch_size, image_size)

        self.patches = flattened_patches
        self.image_size = image_size
        self.attn_head = normalized_ah.tolist()
        self.name = name



def plot_javascript(list_of_attn_heads, list_of_images, list_of_names=None, ATTN_SCALING=8, cls_token=True):

    # if list of attn heads is tensor
    if type(list_of_attn_heads) == torch.Tensor:
        list_of_attn_heads = [np.array(list_of_attn_heads[i]) for i in range(list_of_attn_heads.shape[0])]
    elif type(list_of_attn_heads) != list:
        list_of_attn_heads = [list_of_attn_heads]

    if type(list_of_images) != list:
        list_of_images = [list_of_images]

    if list_of_names == torch.Tensor:
        
        list_of_names = [str(i) for i in list_of_names.tolist()]
    elif list_of_names is None:
        list_of_names = []
        for i in range(len(list_of_attn_heads)):
            list_of_names.append(f"Attention Head {i+1}")


    assert len(list_of_attn_heads) == len(list_of_images), "Must provide an image for each attention head"
    assert len(list_of_attn_heads) == len(list_of_names), "Must provide a name for each attention head"

    attn_head_image_js_infos = []
    for attn_head, image, name in zip(list_of_attn_heads, list_of_images, list_of_names):
        attn_head_image_js_infos.append(AttentionHeadImageJSInfo(attn_head, image, name=name, cls_token=True))

    attn_heads_json = json.dumps([info.attn_head for info in attn_head_image_js_infos])
    patches_json = json.dumps([info.patches for info in attn_head_image_js_infos])
    image_sizes_json = json.dumps([info.image_size for info in attn_head_image_js_infos])
    names_json = json.dumps([info.name for info in attn_head_image_js_infos])

    html_code = generate_html_and_js_code(attn_heads_json, patches_json,image_sizes_json, names_json, ATTN_SCALING, cls_token=cls_token)
    return html_code
    
def generate_html_and_js_code(attn_heads_json, patches_json, image_sizes_json, names_json, ATTN_SCALING, cls_token=True,  canvas_img_id=None, canvas_attn_id=None):

    if canvas_img_id is None:
        canvas_img_id = generate_random_string()
    if canvas_attn_id is None:
        canvas_attn_id = generate_random_string()

    template_folder = os.path.dirname(os.path.abspath(__file__))

    with open(os.path.join(template_folder, 'template.html'), 'r') as file:
        template = Template(file.read())

    # we want all the js in one html so we read all files and pass them in.
    with open(os.path.join(template_folder, 'main_visualize.js'), 'r') as file:
        main_js = file.read()
    with open(os.path.join(template_folder, 'patch_to_img.js'), 'r') as file:
        patch_to_img_js = file.read()
    with open(os.path.join(template_folder, 'get_color.js'), 'r') as file:
        get_color_js = file.read()

    html_code = template.render(
        main_visualize_js = main_js,
        patch_to_img_js = patch_to_img_js,
        get_color_js = get_color_js,

        attn_heads_json=attn_heads_json,
        patches_json=patches_json,
        image_sizes_json=image_sizes_json,
        names_json = names_json,

        ATTN_SCALING=ATTN_SCALING,
        cls_token = cls_token,

        canvas_attn_id = canvas_attn_id,
        canvas_img_id = canvas_img_id
    )

    return html_code


# get layer 0 activations from running the model on the sample image.
def _get_layer_0_activations(model, image):
    layer_0_activations = None

    def hook_fn(m,i,o):
        nonlocal layer_0_activations
        layer_0_activations = i[0][0].cpu().numpy()

    handle = model.blocks[0].attn.attn_scores.register_forward_hook(hook_fn)
    model.eval()
    with torch.no_grad():
        model(torch.from_numpy(np.expand_dims(image,axis=0)).cuda())
    handle.remove()

    return layer_0_activations

def main_example():
    import webbrowser
    import timm
    import copy
    from vit_prisma.utils.get_activations import timmCustomAttention  # Custom Attention has hook functions

    cur_folder = os.path.dirname(os.path.abspath(__file__))

    # load sample image
    sample_image_0 = np.load(os.path.join(cur_folder, "sample_cifar10_image_0.npy"))
    sample_image_10 = np.load(os.path.join(cur_folder, "sample_cifar10_image_10.npy"))

    # Load original model
    orig_model = timm.create_model('vit_base_patch32_224', pretrained=True)

    # Replace original model's attention with hooked attention
    model = copy.deepcopy(orig_model)
    for idx, block in enumerate(model.blocks):
        model.blocks[idx].attn = timmCustomAttention(dim=768, num_heads=12, qkv_bias=True)

    # Reset attention with pretrained weights from original model.
    model.load_state_dict(orig_model.state_dict())
    model = model.cuda()

    layer_0_sample_image_0_activations = _get_layer_0_activations(model, sample_image_0)

    layer_0_sample_image_10_activations = _get_layer_0_activations(model, sample_image_10)
    # plot attention head 3 and 1 ("corner head")

    attn_head_1 = layer_0_sample_image_0_activations[3]
    attn_head_2 = layer_0_sample_image_10_activations[1]

    html_code = plot_javascript([attn_head_1, attn_head_2], [sample_image_0,sample_image_10], list_of_names=["Generic", "Corner"], ATTN_SCALING=8, cls_token=True)

    temp_path = os.path.join(cur_folder, 'temp.html')
    with open(temp_path, 'w') as f:
        f.write(html_code)
    webbrowser.open('file://' + temp_path)

if __name__ == "__main__":
    main_example()



import wandb
import torch
import torch.optim as optim
import tqdm
from tqdm.auto import tqdm
from vit_prisma.training.training_utils import calculate_accuracy, calculate_loss, set_seed, PrismaCallback
from vit_prisma.utils.wandb_utils import dataclass_to_dict, update_dataclass_from_dict
from vit_prisma.training.training_dictionary import optimizer_dict, loss_function_dict
from vit_prisma.training.schedulers import WarmupThenStepLR
from vit_prisma.training.early_stopping import EarlyStopping
from vit_prisma.utils.saving_utils import save_config_to_file
import os
from torch.utils.data import Dataset, DataLoader
import dataclasses
from sklearn.model_selection import train_test_split

def train(
        model_function,
        config,
        train_dataset,
        val_dataset=None,
        checkpoint_path=None,
        callbacks: list[PrismaCallback] = None
):
    if val_dataset is None:
        train_dataset, val_dataset = train_test_split(train_dataset, test_size=0.2)
        print(f"Split train dataset into train and val with {len(train_dataset)} and {len(val_dataset)}.")

    if config.use_wandb:
        if config.wandb_team_name is None:
            wandb.init(project=config.wandb_project_name)
        else:
            wandb.init(entity=config.wandb_team_name, project=config.wandb_project_name)
        sweep_values = wandb.config._items # get sweep values
        update_dataclass_from_dict(config, sweep_values)
        wandb.config.update(dataclass_to_dict(config))
    
    print("Config is:", config)
    save_config_to_file(config, os.path.join(config.parent_dir, "config.json"))

    set_seed(config.seed if config.seed != None else 666)
    model = model_function(config)
    model.train()
    model.to(config.device)
    optimizer = optimizer_dict[config.optimizer_name](model.parameters(), lr = config.lr, weight_decay = config.weight_decay)
    loss_fn = loss_function_dict[config.loss_fn_name]()

    if config.batch_size == -1:
        batch_size_train, batch_size_test = len(train_dataset), len(val_dataset) # use full batch
    else:
        batch_size_train, batch_size_test = config.batch_size, config.batch_size

    train_loader = DataLoader(train_dataset, batch_size=batch_size_train, shuffle=True)
    test_loader = DataLoader(val_dataset, batch_size=batch_size_test, shuffle=False)

    print(f"Length of trainloader {len(train_loader)}.")
    print(f"Length of testloader {len(test_loader)}")
    steps = 0

    scheduler = WarmupThenStepLR(optimizer, warmup_steps=config.warmup_steps, step_size=config.scheduler_step, gamma=config.scheduler_gamma)

    if config.early_stopping:
        early_stopping = EarlyStopping(patience=config.early_stopping_patience, verbose=True)
    else: 
        early_stopping = None
    
    if checkpoint_path and os.path.exists(checkpoint_path):
        checkpoint = torch.load(checkpoint_path)
        model.load_state_dict(checkpoint['model_state_dict'])
        optimizer.load_state_dict(checkpoint['optimizer_state_dict'])
        start_epoch = checkpoint['epoch'] + 1
        print(f"Loaded checkpoint from epoch {checkpoint['epoch']}")
                

    #create dir for checkoint if it doesn't exist
    if os.path.exists(config.parent_dir) and not os.path.exists(os.path.join(config.parent_dir, config.save_dir)):
        os.makedirs(os.path.join(config.parent_dir, config.save_dir))

    num_samples = 0

    for epoch in tqdm(range(1, config.num_epochs + 1)):
        for idx, items in enumerate(train_loader):
            if steps % config.log_frequency == 0:
                model.eval()
                logs = {}
                train_loss = calculate_loss(model, train_loader, loss_fn, config.device)
                test_loss = calculate_loss(model, test_loader, loss_fn, config.device)
                log_dict = {
                    "train_loss": train_loss,
                    "test_loss": test_loss,
                }
                if config.loss_fn_name == "MSE":
                    tqdm.write(f"Steps{steps} | Train loss: {train_loss:.6f} | Test loss: {test_loss:.6f}")
                else:
                    train_acc = calculate_accuracy(model, train_loader, config.device)
                    test_acc = calculate_accuracy(model, test_loader, config.device)
                    tqdm.write(f"Steps{steps} | Train loss: {train_loss:.6f} | Train acc: {train_acc:.5f} | Test loss: {test_loss:.6f} | Test acc: {test_acc:.5f}")
                    log_dict.update({
                                        "train_acc": train_acc, 
                                        "test_acc": test_acc
                                     })
                if config.use_wandb:
                    wandb.log(log_dict, step=num_samples) # Record number of samples
                model.train() # set model back to train mode

            images, labels, *metadata = items
            images, labels = images.to(config.device), labels.to(config.device)

            optimizer.zero_grad()

            y = model(images)
            
            loss = loss_fn(y, labels)
            loss.backward()
            
            torch.nn.utils.clip_grad_norm_(model.parameters(), config.max_grad_norm) if config.max_grad_norm is not None else None
            
            optimizer.step()
            scheduler.step() if config.warmup_steps > 0 else None
            
            tqdm.write(f"Epoch {epoch} | steps{steps} | Num Samples {num_samples} | Loss {loss.item()}") if config.print_every and steps % config.print_every == 0 else None
            
            if config.save_checkpoints and steps % config.save_cp_frequency == 0:
                torch.save({
                    'model_state_dict': model.state_dict(),
                    'optimizer_state_dict': optimizer.state_dict(),
                    'epoch': epoch
                }, os.path.join(os.path.join(config.parent_dir, config.save_dir), f"model_{num_samples}.pth"))
            
            if hasattr(config, 'max_steps') and config.max_steps and steps >= config.max_steps:
                break
            
            steps += 1
            num_samples += len(labels)
            for callback in callbacks:
                callback.on_step_end(steps, model, test_loader, wandb_logger=wandb)

        for callback in callbacks:
            callback.on_epoch_end(epoch, model, test_loader, wandb_logger=wandb)

        if config.early_stopping:
            early_stopping(train_acc)
            if early_stopping.early_stop:
                print("Stopping training due to early stopping!")
                break

    if config.use_wandb:
        wandb.finish()
    return model


import torch
import torch.nn as nn

import einops

from vit_prisma.configs.HookedViTConfig import HookedViTConfig

from typing import Dict, Optional, Tuple, Union

from jaxtyping import Float, Int

class PosEmbedding(nn.Module):

    def __init__(self, cfg: Union[Dict, HookedViTConfig]):

        super().__init__()

        if isinstance(cfg, Dict):
            cfg = HookedViTConfig.from_dict(cfg)
        self.cfg = cfg

        num_patches = (self.cfg.image_size // self.cfg.patch_size)**2
        if self.cfg.is_video_transformer:
            num_patches = num_patches*(self.cfg.video_num_frames//self.cfg.video_tubelet_depth)
            
        token_length = num_patches + 1 if self.cfg.classification_type == 'cls' else num_patches

        self.W_pos = nn.Parameter(
            torch.empty(token_length, self.cfg.d_model, dtype=self.cfg.dtype)
        )
    
    def forward(
            self,
            tokens: Int[torch.Tensor, "batch pos"],
    ):
        pos_embed = self.W_pos
        batch_pos_embed = einops.repeat(pos_embed, "pos d_model -> batch pos d_model", batch=tokens.size(0))
        return batch_pos_embed




import unittest
from vit_prisma.dataloaders.induction import InductionDataset
import torch
import random

SAMPLE_SIZE = 100  # Adjust based on your needs

class TestInductionDataset(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        dir_path = '../data/induction'
        cls.train_dataset = InductionDataset('train', dir_path)
        cls.test_dataset = InductionDataset('test', dir_path)

    def test_dataset_lengths(self):
        # Ensure datasets are not empty
        self.assertTrue(all(len(dataset) > 0 for dataset in [self.train_dataset, self.test_dataset]))

    def _get_sampled_class_counts(self, dataset):
        sampled_indices = random.sample(range(len(dataset)), SAMPLE_SIZE)
        labels = [dataset[i][1] for i in sampled_indices]
        return {label: labels.count(label) for label in set(labels)}

    def test_balanced_classes(self):
        for dataset in [self.train_dataset, self.test_dataset]:
            class_counts = self._get_sampled_class_counts(dataset)
            print("Class counts:", class_counts)

    def test_get_item(self):
        for dataset in [self.train_dataset, self.test_dataset]:
            image, label = dataset[0]
            self.assertEqual(image.shape[0], 1)
            self.assertIsInstance(image, torch.Tensor)

if __name__ == '__main__':
    unittest.main()



import dataclasses

def update_dataclass_from_dict(dc, dct):
    for key, value in dct.items():
        if hasattr(dc, key):
            attr = getattr(dc, key)
            if dataclasses.is_dataclass(attr):
                update_dataclass_from_dict(attr, value)
            else:
                setattr(dc, key, value)

def update_config_with_wandb_sweep(config, sweep_values):
    update_dataclass_from_dict(config, sweep_values)

def dataclass_to_dict(obj):
    if dataclasses.is_dataclass(obj):
        return {name: dataclass_to_dict(value) for name, value in dataclasses.asdict(obj).items()}
    else:
        return obj



import pytest
import torch
from vit_prisma.models.base_vit import HookedViT
import numpy as np
import torch

from transformers import CLIPProcessor, CLIPModel


def test_loading_clip():
    TOLERANCE = 1e-4

    model_name = "wkcn/TinyCLIP-ViT-8M-16-Text-3M-YFCC15M"
    batch_size = 5
    channels = 3
    height = 224
    width = 224
    device = "cpu"


    full = CLIPModel.from_pretrained(model_name)
    tinyclip = full.vision_model

    tinyclip_final_proj = full.visual_projection
    tinyclip.to(device)
    tinyclip_final_proj.to(device)

    hooked_model = HookedViT.from_pretrained(model_name, is_timm=False, is_clip=True)
    hooked_model.to(device)

    with torch.random.fork_rng():
        torch.manual_seed(1)
        input_image = torch.rand((batch_size, channels, height, width)).to(device)
    with torch.no_grad():
        tinyclip_output, hooked_output = tinyclip_final_proj(tinyclip(input_image)[1]),  hooked_model(input_image)

    assert torch.allclose(hooked_output, tinyclip_output, atol=TOLERANCE), f"Model output diverges! Max diff: {torch.max(torch.abs(hooked_output - tinyclip_output))}"





"""
Prisma Repo
By Sonia Joseph

Copyright (c) Sonia Joseph. All rights reserved.

Inspired by TransformerLens. Some functions have been adapted from the TransformerLens project.
For more information on TransformerLens, visit: https://github.com/neelnanda-io/TransformerLens
"""

from typing import Any, Callable, Dict, List, Optional, Tuple, Type, Union, cast
import torch
from jaxtyping import Float, Int
import numpy as np
import re
import json
import logging

from vit_prisma.utils.data_utils.imagenet_dict import IMAGENET_DICT
from vit_prisma.utils.data_utils.imagenet_utils import imagenet_index_from_word


def test_prompt(example_data_point, model, example_answer=None, top_k=10):

    logits = model(example_data_point.unsqueeze(0))
    probs = logits.softmax(dim=-1)
    probs = probs.squeeze(0).detach().numpy()
    sorted_probs = np.sort(probs)[::-1]
    sorted_probs_args = np.argsort(probs)[::-1]

    for i in range(top_k):
        index = sorted_probs_args[i]
        prob = sorted_probs[i]
        logit = logits[0, index].item()  # Assuming you want to show the original logit value
        label = IMAGENET_DICT[index]  # Adjust based on your mapping

        rank_str = f"Top {i}th token."
        logit_str = f"Logit: {logit:.2f}"
        prob_str = f"Prob: {prob * 100:.2f}%"
        token_str = f"Label: |{label}|"

        print(f"{rank_str} {logit_str} {prob_str} {token_str}")

    if example_answer:
      tabby_cat_idx = imagenet_index_from_word(example_answer)

    # Example for displaying ranks of the answer tokens, adjust according to your needs
      answer_index = imagenet_index_from_word(example_answer)
      answer_indices = [answer_index]  # Assuming index is the answer index, adjust as necessary
      print("Rank of the correct answer:")
      for ans_index in answer_indices:
          rank = np.where(sorted_probs_args == ans_index)[0][0]
          print(f"Class Name: {example_answer} | Rank: {rank} | ImageNet Index: {tabby_cat_idx}")

def transpose(tensor: Float[torch.Tensor, "... a b"]) -> Float[torch.Tensor, "... b a"]:
    """
    Utility to swap the last two dimensions of a tensor, regardless of the number of leading dimensions
    """
    return tensor.transpose(-1, -2)

# Type alias
SliceInput: Type = Optional[
    Union[
        int,
        Tuple[int,],
        Tuple[int, int],
        Tuple[int, int, int],
        List[int],
        torch.Tensor,
        np.ndarray,
    ]
]
"""An object that represents a slice input. It can be a tuple of integers or a slice object.

An optional type alias for a slice input used in the `ActivationCache` module.

A `SliceInput` can be one of the following types:
    - `int`: an integer representing a single position
    - `Tuple[int, int]`: a tuple of two integers representing a range of positions
    - `Tuple[int, int, int]`: a tuple of three integers representing a range of positions with a step size
    - `List[int]`: a list of integers representing multiple positions
    - `torch.Tensor`: a tensor containing a boolean mask or a list of indices to be selected from the input tensor.

`SliceInput` is used in the `apply_ln_to_stack` method in the `ActivationCache` module.
"""

class Slice:
    """An object that represents a slice input. It can be a tuple of integers or a slice object.

    We use a custom slice syntax because Python/Torch's don't let us reduce the number of dimensions:

    Note that slicing with input_slice=None means do nothing, NOT add an extra dimension (use unsqueeze for that)

    There are several modes:
    int - just index with that integer (decreases number of dimensions)
    slice - Input is a tuple converted to a slice ((k,) means :k, (k, m) means m:k, (k, m, n) means m:k:n)
    array - Input is a list or tensor or numpy array, converted to a numpy array, and we take the stack of values at those indices
    identity - Input is None, leave it unchanged.

    Examples for dim=0:
    if input_slice=0, tensor -> tensor[0]
    elif input_slice = (1, 5), tensor -> tensor[1:5]
    elif input_slice = (1, 5, 2), tensor -> tensor[1:5:2] (ie indexing with [1, 3])
    elif input_slice = [1, 4, 5], tensor -> tensor[[1, 4, 5]] (ie changing the first axis to have length 3, and taking the indices 1, 4, 5 out).
    elif input_slice is a Tensor, same as list - Tensor is assumed to be a 1D list of indices.
    """

    def __init__(
        self,
        input_slice: SliceInput = None,
    ):
        """
        Modular component for slicing tensors. Can be used to slice a tensor along a given dimension, or to index into a tensor along a given dimension.

        Args:
            input_slice (SliceInput): The slice to apply. Can be an int, a tuple, a list, a torch.Tensor, or None. If None, do nothing.

        Raises:
            ValueError: If the input_slice is not one of the above types.
        """
        if type(input_slice) == tuple:
            input_slice: slice = slice(*input_slice)
            self.slice = input_slice
            self.mode = "slice"
        elif type(input_slice) == int:
            self.slice = input_slice
            self.mode = "int"
        elif type(input_slice) == slice:
            self.slice = input_slice
            self.mode = "slice"
        elif type(input_slice) in [list, torch.Tensor, np.ndarray]:
            self.slice = to_numpy(input_slice)
            self.mode = "array"
        elif input_slice is None:
            self.slice = slice(None)
            self.mode = "identity"
        else:
            raise ValueError(f"Invalid input_slice {input_slice}")
        

    def apply(
        self,
        tensor: torch.Tensor,
        dim: int = 0,
    ) -> torch.Tensor:
        """
        Takes in a tensor and a slice, and applies the slice to the given dimension (supports positive and negative dimension syntax). Returns the sliced tensor.

        Args:
            tensor (torch.Tensor): The tensor to slice.
            dim (int, optional): The dimension to slice along. Supports positive and negative dimension syntax.

        Returns:
            torch.Tensor: The sliced tensor.
        """
        ndim = tensor.ndim
        slices = [slice(None)] * ndim
        slices[dim] = self.slice
        return tensor[tuple(slices)]

    def indices(
        self,
        max_ctx: Optional[int] = None,
    ) -> Union[np.ndarray, np.int32, np.int64]:
        """
        Returns the indices when this slice is applied to an axis of size max_ctx. Returns them as a numpy array, for integer slicing it is eg array([4])

        Args:
            max_ctx (int, optional): The size of the axis to slice. Only used if the slice is not an integer.

        Returns:
            np.ndarray: The indices that this slice will select.

        Raises:
            ValueError: If the slice is not an integer and max_ctx is not specified.
        """
        if self.mode == "int":
            return np.array([self.slice], dtype=np.int64)
        if max_ctx is None:
            raise ValueError("max_ctx must be specified if slice is not an integer")
        return np.arange(max_ctx, dtype=np.int64)[self.slice]

    def __repr__(
        self,
    ) -> str:
        return f"Slice: {self.slice} Mode: {self.mode} "
    


def get_act_name(
    name: str,
    layer: Optional[Union[int, str]] = None,
    layer_type: Optional[str] = None,
):
    """
    Helper function to convert shorthand to an activation name. Pretty hacky, intended to be useful for short feedback
    loop hacking stuff together, more so than writing good, readable code. But it is deterministic!

    Returns a name corresponding to an activation point in a TransformerLens model.

    Args:
         name (str): Takes in the name of the activation. This can be used to specify any activation name by itself.
         The code assumes the first sequence of digits passed to it (if any) is the layer number, and anything after
         that is the layer type.

         Given only a word and number, it leaves layer_type as is.
         Given only a word, it leaves layer and layer_type as is.

         Examples:
             get_act_name('embed') = get_act_name('embed', None, None)
             get_act_name('k6') = get_act_name('k', 6, None)
             get_act_name('scale4ln1') = get_act_name('scale', 4, 'ln1')

         layer (int, optional): Takes in the layer number. Used for activations that appear in every block.

         layer_type (string, optional): Used to distinguish between activations that appear multiple times in one block.

    Full Examples:

    get_act_name('k', 6, 'a')=='blocks.6.attn.hook_k'
    get_act_name('pre', 2)=='blocks.2.mlp.hook_pre'
    get_act_name('embed')=='hook_embed'
    get_act_name('normalized', 27, 'ln2')=='blocks.27.ln2.hook_normalized'
    get_act_name('k6')=='blocks.6.attn.hook_k'
    get_act_name('scale4ln1')=='blocks.4.ln1.hook_scale'
    get_act_name('pre5')=='blocks.5.mlp.hook_pre'
    """
    if (
        ("." in name or name.startswith("hook_"))
        and layer is None
        and layer_type is None
    ):
        # If this was called on a full name, just return it
        return name
    match = re.match(r"([a-z]+)(\d+)([a-z]?.*)", name)
    if match is not None:
        name, layer, layer_type = match.groups(0)

    layer_type_alias = {
        "a": "attn",
        "m": "mlp",
        "b": "",
        "block": "",
        "blocks": "",
        "attention": "attn",
    }

    act_name_alias = {
        "attn": "pattern",
        "attn_logits": "attn_scores",
        "key": "k",
        "query": "q",
        "value": "v",
        "mlp_pre": "pre",
        "mlp_mid": "mid",
        "mlp_post": "post",
    }

    layer_norm_names = ["scale", "normalized"]

    if name in act_name_alias:
        name = act_name_alias[name]

    full_act_name = ""
    if layer is not None:
        full_act_name += f"blocks.{layer}."
    if name in [
        "k",
        "v",
        "q",
        "z",
        "rot_k",
        "rot_q",
        "result",
        "pattern",
        "attn_scores",
    ]:
        layer_type = "attn"
    elif name in ["pre", "post", "mid", "pre_linear"]:
        layer_type = "mlp"
    elif layer_type in layer_type_alias:
        layer_type = layer_type_alias[layer_type]

    if layer_type:
        full_act_name += f"{layer_type}."
    full_act_name += f"hook_{name}"

    if name in layer_norm_names and layer is None:
        full_act_name = f"ln_final.{full_act_name}"
    return full_act_name

def to_numpy(tensor):
    """
    Helper function to convert a tensor to a numpy array. Also works on lists, tuples, and numpy arrays.
    """
    if isinstance(tensor, np.ndarray):
        return tensor
    elif isinstance(tensor, (list, tuple)):
        array = np.array(tensor)
        return array
    elif isinstance(tensor, (torch.Tensor, torch.nn.parameter.Parameter)):
        return tensor.detach().cpu().numpy()
    elif isinstance(tensor, (int, float, bool, str)):
        return np.array(tensor)
    else:
        raise ValueError(f"Input to to_numpy has invalid type: {type(tensor)}")




import torch.nn as nn
import torch

from typing import Dict, Union

from jaxtyping import Float

from vit_prisma.configs.HookedViTConfig import HookedViTConfig
from fancy_einsum import einsum

# maps to class logits
class Head(nn.Module):
    def __init__(self, cfg: Union[Dict, HookedViTConfig]):
        super().__init__()
        if isinstance(cfg, Dict):
            cfg = HookedViTConfig.from_dict(cfg)
        self.cfg = cfg

        self.W_H: Float[torch.Tensor, "d_model n_classes"] = nn.Parameter(
            torch.empty(self.cfg.d_model, self.cfg.n_classes, dtype=cfg.dtype)
        )
        self.b_H: Float[torch.Tensor, "n_classes"] = nn.Parameter(
            torch.zeros(self.cfg.n_classes, dtype=cfg.dtype)
        )

    def forward(
        self, residual: Float[torch.Tensor, "batch d_model"]
    ) -> Float[torch.Tensor, "batch n_classes"]:
        
        return (
            einsum(
                "batch d_model, d_model n_classes -> batch n_classes",
                residual,
                self.W_H,
            )
            + self.b_H
        )


import torch
from tqdm import tqdm
import torch.nn as nn
import numpy as np

def get_activations(net,layer,data_loader,use_cuda=True, max_count=0, test_run=False, debug=True):
    '''
    Get activations of a layer for a given data_loader
    '''
    activations = []
    def hook_fn(m,i,o):
        activations.append(i[0].cpu().numpy())
    handle = layer.register_forward_hook(hook_fn)
    if use_cuda:
        net = net.cuda()
    net.eval()

    count = 0
    for i, (images, labels) in enumerate(tqdm(data_loader)):
        # if debug: print(i)
        if use_cuda:
            images = images.cuda()
        with torch.no_grad():
            output = net(images)

        count += 1
        if count > max_count:
            break
    handle.remove()
    activations_np = np.vstack(activations)     # assuming first dimension is num_examples: batches x batch_size x <feat_dims> --> num_examples x <feat_dims>
    return activations_np

class ActivationCacheHook:
    '''
    Hook to save activations of a layer
    '''
    def __init__(self):
        self.activations = []

    def __call__(self, module, input, output):
        self.activations.append(output.clone())

class CustomAttention(nn.Module):

    def __init__(self, config, logger = None):
        super().__init__()

        self.logger = logger 
        self.config = config

        hidden_dim = self.config.transformer.hidden_dim

        assert hidden_dim % self.config.transformer.num_heads == 0, "Embedding dimension must be divisible by number of heads"
        self.head_dim = hidden_dim // self.config.transformer.num_heads
        self.scale = self.head_dim ** -0.5
        self.qkv = nn.Linear(hidden_dim, 3 * hidden_dim, bias=False)
        self.q_norm, self.k_norm = nn.LayerNorm(self.head_dim) if self.config.layernorm.qknorm else nn.Identity(), nn.LayerNorm(self.head_dim) if self.config.layernorm.qknorm else nn.Identity()
        self.attn_dropout = nn.Dropout(self.config.dropout.attention) if self.config.dropout.attention > 0 else nn.Identity()
        self.proj = nn.Linear(hidden_dim, hidden_dim) if self.config.transformer.attn_hidden_layer else nn.Identity()
        self.proj_dropout = nn.Dropout(self.config.dropout.proj) if self.config.dropout.proj > 0 else nn.Identity()
        
        self._log(f"Attention layer initialized with config {self.config}")

        self.attn_scores = nn.Identity()
        self.attn_pattern = nn.Identity()


    def _log(self, msg):
        if self.logger:
            self.logger.info(msg)

    def forward(self, x):
        self._log(f"Attention input size is {x.shape}") 
        B, N, C = x.shape
        qkv = self.qkv(x).reshape(B, N, 3, self.config.transformer.num_heads, self.head_dim).permute(2, 0, 3, 1, 4)
        q, k, v = qkv.unbind(0)
        q, k = self.q_norm(q), self.k_norm(k)
        q = q * self.scale
        attn = q @ k.transpose(-2, -1)
        attention_scores = self.attn_scores(attn)
        attn = attn.softmax(dim=-1)
        attn_pattern  = self.attn_pattern(attn)
        self._log(f"Attention size after softmax is {attn.shape}") 
        attn = self.attn_dropout(attn)
        x = (attn @ v).transpose(1, 2).reshape(B, N, C)
        x = self.proj(x)
        x = self.proj_dropout(x)
        self._log(f"Attention output size is {x.shape}") 
        return x

def get_activations(net,layer,data_loader,use_cuda=True, max_count=0, test_run=False, debug=True):
    '''
    Get activations of a layer for a given data_loader
    '''
    activations = []
    def hook_fn(m,i,o):
        activations.append(i[0].cpu().numpy())
    handle = layer.register_forward_hook(hook_fn)
    if use_cuda:
        net = net.cuda()
    net.eval()

    count = 0
    for i, (images, labels) in enumerate(tqdm(data_loader)):
        # if debug: print(i)
        if use_cuda:
            images = images.cuda()
        with torch.no_grad():
            output = net(images)

        count += 1
        if count > max_count:
            break
    handle.remove()
    activations_np = np.vstack(activations)     # assuming first dimension is num_examples: batches x batch_size x <feat_dims> --> num_examples x <feat_dims>
    return activations_np


class timmCustomAttention(nn.Module):
    '''
    Custom attention layer for timm models.
    '''

    def __init__(
            self,
            dim,
            num_heads=8,
            qkv_bias=False,
            qk_norm=False,
            attn_drop=0.,
            proj_drop=0.,
            norm_layer=nn.LayerNorm,
    ):
        super().__init__()
        assert dim % num_heads == 0, 'dim should be divisible by num_heads'
        self.num_heads = num_heads
        self.head_dim = dim // num_heads
        self.scale = self.head_dim ** -0.5

        self.qkv = nn.Linear(dim, dim * 3, bias=qkv_bias)
        self.q_norm = norm_layer(self.head_dim) if qk_norm else nn.Identity()
        self.k_norm = norm_layer(self.head_dim) if qk_norm else nn.Identity()
        self.attn_drop = nn.Dropout(attn_drop)
        self.proj = nn.Linear(dim, dim)
        self.proj_drop = nn.Dropout(proj_drop)

        self.attn_scores = nn.Identity()
        self.attn_pattern = nn.Identity()  # Initialize attn_pattern


    def forward(self, x):
        B, N, C = x.shape
        qkv = self.qkv(x).reshape(B, N, 3, self.num_heads, self.head_dim).permute(2, 0, 3, 1, 4)
        q, k, v = qkv.unbind(0)
        q, k = self.q_norm(q), self.k_norm(k)

        q = q * self.scale

        attn_scores = q @ k.transpose(-2, -1)
        attn_scores = self.attn_scores(attn_scores)

        attn_pattern = attn_scores.softmax(dim=-1)
        attn_pattern = self.attn_pattern(attn_pattern) # For hook function

        attn = self.attn_drop(attn_pattern)

        x = attn @ v

        x = x.transpose(1, 2).reshape(B, N, C)
        x = self.proj(x)
        x = self.proj_drop(x)
        return x


import random
import numpy as np
import os
from sklearn.model_selection import train_test_split
from torch.utils.data import Dataset
import torch
from vit_prisma.dataloaders.induction import draw_circle, draw_line, draw_x, draw_diagonal

class PolygenicInductionDataset(Dataset):
    def __init__(self, train_or_test, dir_path = '../data/polygenic_induction', use_metadata=False, transform=None):

        self.dir_path = dir_path
        self.cache_path = f'{dir_path}/all_{train_or_test}.npz'

        self.use_metadata = use_metadata
        self.transform = transform

        if not os.path.exists(self.cache_path):
            print("Generating and saving new polygenic induction dataset...")
            self._generate_and_cache()

        print("Loading induction dataset from cache...", self.cache_path)
        self._load_from_cache()
        # self._normalize_and_to_tensor()

    def __len__(self):
        return len(self.images)

    def __getitem__(self, idx):
        image = self.images[idx][np.newaxis, :, :]
        label = self.labels[idx]
        # meta_data = self.metadata[idx]
        image = torch.from_numpy(image).float()
        if self.transform:
            image = self.transform(image)
        return image, label

    def _load_from_cache(self):
        loaded = np.load(self.cache_path, allow_pickle=True, mmap_mode='r' )
        self.images = loaded['images']
        if self.use_metadata:
            self.metadata = loaded['metadata']
        self.labels = loaded['labels']

    def _generate_and_cache(self):
        generate_dataset(self.dir_path)

    # def _normalize_and_to_tensor(self):
    #     self.images = 2 * torch.tensor(self.images, dtype=torch.float32) / 255.0 - 1 # normalize [-1,]
    #     self.labels = torch.tensor(self.labels, dtype=torch.int64)

## Functionality to generate dataset ##

def plot_four_objects(A, B, C, D, Ax, Ay, Bx, By, Cx, Cy, Dx, Dy, vertical=False):  

    image = np.zeros((64, 64))
    
    A(image, Ax, Ay, im_size=64)
    B(image, Bx, By, im_size=64)
    C(image, Cx, Cy, im_size=64)
    D(image, Dx, Dy, im_size=64)

    if vertical:
        image = image.T
    return image

def generate_dataset(dir_path = '../data/polygenic_induction'):
    draw_functions = [draw_circle, draw_line, draw_x, draw_diagonal]
    padding = 4
    offset = 7
    max_shape_size = 5  

    images = []
    metadata = []
    labels = [] 

    max_a = 64 - 3 * offset - 2 * (padding + max_shape_size)
    max_b = 64 - padding - max_shape_size

    arrangements = {
        'A A A A' : 0,
        'A B A B' : 1,
        'A B B A' : 2,
        'A A B B' : 3,
        'A B B B' : 4,
        'A A A B' : 5,
    }

    for vertical in [True, False]:
        for a in range(padding + max_shape_size, max_a):
            for b in range(padding + max_shape_size, max_b):
                Ax = a
                Ay = b
                Bx = Ax + offset
                By = Ay
                Cx = Bx + offset
                Cy = Ay
                Dx = Cx + offset
                Dy = By

                for A in draw_functions:
                    for B in draw_functions:
                        for arrangement in arrangements.keys():

                            text_label = arrangement
                            l = arrangements[arrangement] + (0 if vertical else 6)
                            arrangement = arrangement.split()

                            ta = locals()[arrangement[0]]
                            tb = locals()[arrangement[1]]
                            tc = locals()[arrangement[2]]
                            td = locals()[arrangement[3]]

                            if len(set(text_label.split())) != len(set([ta, tb, tc, td])):
                                continue

                            img = plot_four_objects(ta, tb, tc, td, Ax, Ay, Bx, By, Cx, Cy, Dx, Dy, vertical=vertical)

                            if ta == tb == tc == td:
                                all_same = True
                            else:
                                all_same = False

                            images.append(img)

                            m = {
                                "Ax": Ax,
                                "Ay": Ay,
                                "Bx": Bx,
                                "By": By,
                                "Cx": Cx,
                                "Cy": Cy,
                                "Dx": Dx,
                                "Dy": Dy,
                                "A": ta.__name__,
                                "B": tb.__name__,
                                "C": tc.__name__,
                                "D": td.__name__,
                                "Same": all_same,
                                "Vertical": vertical,
                                "pattern": text_label,
                            }

                            metadata.append(m)
                            labels.append(l)

    path = f'{dir_path}/induction_dataset.npz'
    if not os.path.exists(dir_path):
        os.makedirs(dir_path)
    np.savez(path, images=images, metadata=metadata, labels=labels)

    print("Balancing dataset...")
    create_balanced_dataset(dir_path)


def create_balanced_dataset(dir_path='../data/polygenic_induction'):
    
    # Load the dataset
    dataset = np.load(f'{dir_path}/induction_dataset.npz', mmap_mode='r', allow_pickle=True)
    images, metadata, labels = dataset['images'], dataset['metadata'], dataset['labels']

    # Create categories based on the conditions
    categories = {}

    for label in range(0, 12):
        categories[label] = [i for i, entry in enumerate(labels) if entry == label]

    sample_size = min(len(c) for c in categories.values())

    # Sample from each category and save as separate npz files
    for key, val in categories.items():
        indices = random.sample(val, sample_size)
        np.savez(f'{dir_path}/{key}.npz', 
                 images=images[indices], 
                 metadata=metadata[indices], 
                 labels=labels[indices])

    # Split each npz file into train/test datasets
    all_train_data, all_test_data = [], []
    for key in categories.keys():
        data = np.load(f'{dir_path}/{key}.npz', allow_pickle=True)
        train, test = train_test_split(list(zip(data['images'], data['metadata'], data['labels'])), test_size=0.1, random_state=42)
        all_train_data.extend(train)
        all_test_data.extend(test)

    # Shuffle and save combined train/test datasets
    random.shuffle(all_train_data)
    random.shuffle(all_test_data)

    train_images, train_metadata, train_labels = zip(*all_train_data)
    test_images, test_metadata, test_labels = zip(*all_test_data)

    np.savez(f'{dir_path}/all_train.npz', images=np.array(train_images), metadata=np.array(train_metadata), labels=np.array(train_labels))
    np.savez(f'{dir_path}/all_test.npz', images=np.array(test_images), metadata=np.array(test_metadata), labels=np.array(test_labels))
    print("Saved total train and test.")



import torch
import numpy as np
import matplotlib.pyplot as plt

def display_grid_on_image(image, patch_size=32):

    if isinstance(image, torch.Tensor):
        image = image.detach().numpy().transpose(1,2,0)

    if image.shape[0] != 224:
        image = image.transpose(1,2,0)

    if image.max() <= 1.0:
        image = (image * 255).astype(np.uint8)

    num_patches = (image.shape[0] / patch_size) ** 2

    grid_size = int(np.sqrt(num_patches))

    # Calculate patch size
    patch_height = image.shape[0] // grid_size
    patch_width = image.shape[1] // grid_size

    # Overlay grida
    grid_image = np.copy(image)
    for i in range(1, grid_size):
        # Vertical lines
        grid_image[:, patch_width * i, :] = [255, 255, 255]
        # Horizontal lines
        grid_image[patch_height * i, :, :] = [255, 255, 255]

    plt.figure(figsize=(4, 4), dpi=100)  # Adjust figsize and dpi as needed

    # Place labels
    for i in range(grid_size):
        for j in range(grid_size):
            x_center = (j * patch_width) + (patch_width // 2)
            y_center = (i * patch_height) + (patch_height // 2)
            # Convert the patch index to 2D coordinates (row, column)
            patch_index = i * grid_size + j
            row, col = divmod(patch_index, grid_size)
            plt.text(x_center, y_center, f"{patch_index+1}", color='red', fontsize=8, ha='center', va='center')

    # Display image with grid and labels
    # plt.figure(figsize=(8, 8))
    plt.imshow(grid_image)
    plt.axis('off')
    plt.show()



import pytest
import torch
import timm
from vit_prisma.models.base_vit import HookedViT

#currently only vit_base_patch16_224 supported (config loading issue)
def test_loading_timm():
    TOLERANCE = 1e-5

    model_name = "vit_base_patch16_224"
    batch_size = 5
    channels = 3
    height = 224
    width = 224
    device = "cpu"

    hooked_model = HookedViT.from_pretrained(model_name)
    hooked_model.to(device)
    timm_model = timm.create_model(model_name, pretrained=True)
    timm_model.to(device)

    with torch.random.fork_rng():
        torch.manual_seed(1)
        input_image = torch.rand((batch_size, channels, height, width)).to(device)

    hooked_output, timm_output = hooked_model(input_image), timm_model(input_image)

    assert torch.allclose(hooked_output, timm_output, atol=TOLERANCE), f"Model output diverges! Max diff: {torch.max(torch.abs(hooked_output - timm_output))}"



import torch

class WarmupThenStepLR(torch.optim.lr_scheduler._LRScheduler):
    def __init__(self, optimizer, warmup_steps, step_size, gamma=0.5, last_epoch=-1):
        # Does warmup for warmup steps, then moves to step decay parameterized by step_size
        self.warmup_steps = warmup_steps
        self.step_size = step_size
        self.gamma = gamma
        super(WarmupThenStepLR, self).__init__(optimizer, last_epoch)

    def get_lr(self):
        if self.last_epoch < self.warmup_steps:
            return [base_lr * (self.last_epoch / self.warmup_steps) for base_lr in self.base_lrs]
        else:
            # Check if it's time for step decay
            if (self.last_epoch - self.warmup_steps + 1) % self.step_size == 0:
                current_lr = self.optimizer.param_groups[0]['lr']  # Assuming all parameter groups have the same learning rate
                print(f"Reducing learning rate from {current_lr} to {current_lr * self.gamma} at epoch {self.last_epoch}")

            return [base_lr * (self.gamma ** ((self.last_epoch - self.warmup_steps) // self.step_size))
                    for base_lr in self.base_lrs]


"""
Prisma Repo
By Sonia Joseph

Copyright (c) Sonia Joseph. All rights reserved.

Inspired by TransformerLens. Some functions have been adapted from the TransformerLens project.
For more information on TransformerLens, visit: https://github.com/neelnanda-io/TransformerLens
"""

import logging

import torch
import torch.nn as nn

from transformers import ViTForImageClassification

from vit_prisma.models.layers.patch_embedding import PatchEmbedding, TubeletEmbedding
from vit_prisma.models.layers.position_embedding import PosEmbedding
from vit_prisma.models.layers.layer_norm import LayerNorm, LayerNormPre
from vit_prisma.models.layers.mlp import MLP
from vit_prisma.models.layers.attention import Attention
from vit_prisma.models.layers.transformer_block import TransformerBlock, BertBlock
from vit_prisma.models.layers.head import Head

from vit_prisma.training.training_dictionary import activation_dict, initialization_dict
# from vit_prisma.models.prisma_net import PrismaNet
from vit_prisma.prisma_tools.hook_point import HookPoint
from vit_prisma.prisma_tools.hooked_root_module import HookedRootModule

from vit_prisma.configs import HookedViTConfig

from vit_prisma.prisma_tools.activation_cache import ActivationCache

from vit_prisma.prisma_tools.loading_from_pretrained import convert_pretrained_model_config, get_pretrained_state_dict, fill_missing_keys
from vit_prisma.utils.prisma_utils import transpose

from vit_prisma.utils import devices
from vit_prisma.prisma_tools import FactoredMatrix

from typing import Union, Dict, List, Tuple, Optional, Literal

from jaxtyping import Float, Int

import einops
from fancy_einsum import einsum

DTYPE_FROM_STRING = {
    "float32": torch.float32,
    "fp32": torch.float32,
    "float16": torch.float16,
    "fp16": torch.float16,
    "bfloat16": torch.bfloat16,
    "bf16": torch.bfloat16,
}

class HookedViT(HookedRootModule):
    """
    Base vision model.
    Based on 'An Image is Worth 16x16 Words: Transformers for Image Recognition at Scale' https://arxiv.org/abs/2010.11929.
    Adapted from TransformerLens: https://github.com/neelnanda-io/TransformerLens/blob/main/transformer_lens/HookedTransformer.py
    Inspiration also taken from the timm library.
    """

    def __init__(
            self,
            cfg: HookedViTConfig,
            move_to_device: bool = True,
    ):
        """
        Model initialization
        """

        super().__init__()
        if isinstance(cfg, Dict):
            cfg = HookedViTConfig(**cfg)
        elif isinstance(cfg, str):
            raise ValueError(
                "Please pass in a config dictionary or HookedViT object. If you want to load a "
                "pretrained model, use HookedViT.from_pretrained() instead."
            )
        self.cfg = cfg

        # ClS token
        self.cls_token = nn.Parameter(torch.randn(1, 1, self.cfg.d_model))

        # Patch embeddings
        if self.cfg.is_video_transformer:
            self.embed = TubeletEmbedding(self.cfg)
        else:
            self.embed = PatchEmbedding(self.cfg)
        self.hook_embed = HookPoint()

        # Position embeddings
        self.pos_embed = PosEmbedding(self.cfg)
        self.hook_pos_embed = HookPoint()

        if self.cfg.layer_norm_pre: # Put layernorm after attn/mlp layers, not before
            if self.cfg.normalization_type == "LN":
                self.ln_pre = LayerNorm(self.cfg)
            elif self.cfg.normalization_type == "LNPre":
                self.ln_pre = LayerNormPre(self.cfg)
            elif self.cfg.normalization_type is None:
                self.ln_pre = nn.Identity()
            else:
                raise ValueError(f"Invalid normalization type: {self.cfg.normalization_type}")
            self.hook_ln_pre = HookPoint()

        # Blocks
        if self.cfg.use_bert_block:
            block = BertBlock
        else:
            block = TransformerBlock
        
        self.blocks = nn.ModuleList(
            [
                block(self.cfg, block_index)
                for block_index in range(self.cfg.n_layers)
            ]
        )
        # Final layer norm
        if self.cfg.normalization_type == "LN":
            self.ln_final = LayerNorm(self.cfg)
        elif self.cfg.normalization_type == "LNPre":
            self.ln_final = LayerNormPre(self.cfg)
        elif self.cfg.normalization_type is None:
            self.ln_final = nn.Identity()
        else:
            raise ValueError(f"Invalid normalization type: {self.cfg.normalization_type}")

        # Final classification head
        self.head = Head(self.cfg)


        # Initialize weights
        self.init_weights()



        # Set up HookPoints
        self.setup()

    def forward(self,
            input: Union[
            Float[torch.Tensor, "batch height width channels"],

            ],
            stop_at_layer: Optional[int] = None,

        ):
        """Forward Pass.
        Args:
            stop_at_layer Optional[int]: If not None, stop the forward pass at the specified layer.
                Exclusive - ie, stop_at_layer = 0 will only run the embedding layer, stop_at_layer =
                1 will run the embedding layer and the first transformer block, etc. Supports
                negative indexing. Useful for analysis of intermediate layers, eg finding neuron
                activations in layer 3 of a 24 layer model. Defaults to None (run the full model).
                If not None, we return the last residual stream computed.
        """



        batch_size = input.shape[0]

        embed = self.hook_embed(self.embed(input))

        if self.cfg.classification_type == 'cls':
            cls_tokens = self.cls_token.expand(batch_size, -1, -1)  # CLS token for each item in the batch
            embed = torch.cat((cls_tokens, embed), dim=1) # Add to embedding
        pos_embed = self.hook_pos_embed(self.pos_embed(input))
        
        residual = embed + pos_embed

        if self.cfg.layer_norm_pre:
            residual = self.ln_pre(residual)
            residual = self.hook_ln_pre(residual)

        for block in self.blocks[:stop_at_layer]:
            residual = block(residual)
        if stop_at_layer is not None:
            return residual

        x = self.ln_final(residual)

        if self.cfg.classification_type == 'gaap':  # GAAP
            x = x.mean(dim=1)
        elif self.cfg.classification_type == 'cls':  # CLS token
            x = x[:, 0]
            
        return x if self.cfg.return_type == 'pre_logits' else self.head(x)

    def init_weights(self):
        if self.cfg.classification_type == 'cls':
            nn.init.normal_(self.cls_token, std=self.cfg.cls_std)
        # nn.init.trunc_normal_(self.position_embedding, std=self.cfg.pos_std)   
        if self.cfg.weight_type == 'he':
            for m in self.modules(): 
                if isinstance(m, PosEmbedding):
                    nn.init.normal_(m.W_pos, std=self.cfg.pos_std)
                elif isinstance(m, nn.Linear) or isinstance(m, nn.Conv2d):
                    nn.init.kaiming_normal_(m.weight, nonlinearity='relu')
                    if m.bias is not None:
                        nn.init.constant_(m.bias, 0)

    def run_with_cache(
        self, *model_args, return_cache_object=True, remove_batch_dim=False, **kwargs
    ) -> Tuple[
        Union[
            None,
            Float[torch.Tensor, "batch n_classes"],
        ],
        Union[ActivationCache, Dict[str, torch.Tensor]],
    ]:
        """Wrapper around `run_with_cache` in HookedRootModule.

        If return_cache_object is True, this will return an ActivationCache object, with a bunch of
        useful HookedTransformer specific methods, otherwise it will return a dictionary of
        activations as in HookedRootModule.
        """
        out, cache_dict = super().run_with_cache(
            *model_args, remove_batch_dim=remove_batch_dim, **kwargs
        )
        if return_cache_object:
            cache = ActivationCache(
                cache_dict, self, has_batch_dim=not remove_batch_dim
            )
            return out, cache
        else:
            return out, cache_dict
        
    def tokens_to_residual_directions(self, labels):
        '''
        Logit-lens related funtions not implemented; see how we can implement a vision equivalent.
        '''

        answer_residual_directions = self.head.W_H[:,labels]  
        answer_residual_directions = einops.rearrange(
                        answer_residual_directions, "d_model ... -> ... d_model"
                    )
        
        return answer_residual_directions

    def fold_layer_norm(
        self, state_dict: Dict[str, torch.Tensor], fold_biases=True, center_weights=True
    ):
        """Fold Layer Norm. Can also be used to fold RMS Norm, when fold_biases and center_weights are set to False.

        Takes in a state dict from a pretrained model, formatted to be consistent with
        HookedTransformer but with LayerNorm weights and biases. Folds these into the neighbouring
        weights. See further_comments.md for more details.

        Args:
            state_dict (Dict[str, torch.Tensor]): State dict of pretrained model.
            fold_biases (bool): Enables folding of LN biases. Should be disabled when RMS Norm is used.
            center_weights (bool): Enables the centering of weights after folding in LN. Should be disabled when RMS Norm is used.
        """

        # Models that use Grouped Query Attention (Only Mistral at the time of writing) prefix their K/V weights and
        # biases with an underscore in order to distinguish them, but folding the LN into them still works the same,
        # so we just add the underscore if GQA is used (i.e. if `cfg.n_key_value_heads is specified`).
        gqa = "" if self.cfg.n_key_value_heads is None else "_"

        for l in range(self.cfg.n_layers):
            # Fold ln1 into attention - it's important to fold biases first, since biases depend on
            # weights but not vice versa The various indexing is just to broadcast ln.b and ln.w
            # along every axis other than d_model. Each weight matrix right multiplies. To fold in
            # the bias, we use the W_ matrix to map it to the hidden space of the layer, so we need
            # to sum along axis -2, which is the residual stream space axis.
            if fold_biases:
                state_dict[f"blocks.{l}.attn.b_Q"] = state_dict[
                    f"blocks.{l}.attn.b_Q"
                ] + (
                    state_dict[f"blocks.{l}.attn.W_Q"]
                    * state_dict[f"blocks.{l}.ln1.b"][None, :, None]
                ).sum(
                    -2
                )
                state_dict[f"blocks.{l}.attn.{gqa}b_K"] = state_dict[
                    f"blocks.{l}.attn.{gqa}b_K"
                ] + (
                    state_dict[f"blocks.{l}.attn.{gqa}W_K"]
                    * state_dict[f"blocks.{l}.ln1.b"][None, :, None]
                ).sum(
                    -2
                )
                state_dict[f"blocks.{l}.attn.{gqa}b_V"] = state_dict[
                    f"blocks.{l}.attn.{gqa}b_V"
                ] + (
                    state_dict[f"blocks.{l}.attn.{gqa}W_V"]
                    * state_dict[f"blocks.{l}.ln1.b"][None, :, None]
                ).sum(
                    -2
                )
                del state_dict[f"blocks.{l}.ln1.b"]

            state_dict[f"blocks.{l}.attn.W_Q"] = (
                state_dict[f"blocks.{l}.attn.W_Q"]
                * state_dict[f"blocks.{l}.ln1.w"][None, :, None]
            )
            state_dict[f"blocks.{l}.attn.{gqa}W_K"] = (
                state_dict[f"blocks.{l}.attn.{gqa}W_K"]
                * state_dict[f"blocks.{l}.ln1.w"][None, :, None]
            )
            state_dict[f"blocks.{l}.attn.{gqa}W_V"] = (
                state_dict[f"blocks.{l}.attn.{gqa}W_V"]
                * state_dict[f"blocks.{l}.ln1.w"][None, :, None]
            )
            del state_dict[f"blocks.{l}.ln1.w"]

            # Finally, we center the weights reading from the residual stream. The output of the
            # first part of the LayerNorm is mean 0 and standard deviation 1, so the mean of any
            # input vector of the matrix doesn't matter and can be set to zero. Equivalently, the
            # output of LayerNormPre is orthogonal to the vector of all 1s (because dotting with
            # that gets the sum), so we can remove the component of the matrix parallel to this.
            if center_weights:
                state_dict[f"blocks.{l}.attn.W_Q"] -= einops.reduce(
                    state_dict[f"blocks.{l}.attn.W_Q"],
                    "head_index d_model d_head -> head_index 1 d_head",
                    "mean",
                )
                state_dict[f"blocks.{l}.attn.{gqa}W_K"] -= einops.reduce(
                    state_dict[f"blocks.{l}.attn.{gqa}W_K"],
                    "head_index d_model d_head -> head_index 1 d_head",
                    "mean",
                )
                state_dict[f"blocks.{l}.attn.{gqa}W_V"] -= einops.reduce(
                    state_dict[f"blocks.{l}.attn.{gqa}W_V"],
                    "head_index d_model d_head -> head_index 1 d_head",
                    "mean",
                )

            # Fold ln2 into MLP
            if not self.cfg.attn_only:
                if fold_biases:
                    state_dict[f"blocks.{l}.mlp.b_in"] = state_dict[
                        f"blocks.{l}.mlp.b_in"
                    ] + (
                        state_dict[f"blocks.{l}.mlp.W_in"]
                        * state_dict[f"blocks.{l}.ln2.b"][:, None]
                    ).sum(
                        -2
                    )
                    del state_dict[f"blocks.{l}.ln2.b"]

                state_dict[f"blocks.{l}.mlp.W_in"] = (
                    state_dict[f"blocks.{l}.mlp.W_in"]
                    * state_dict[f"blocks.{l}.ln2.w"][:, None]
                )

                if self.cfg.gated_mlp:
                    state_dict[f"blocks.{l}.mlp.W_gate"] = (
                        state_dict[f"blocks.{l}.mlp.W_gate"]
                        * state_dict[f"blocks.{l}.ln2.w"][:, None]
                    )

                del state_dict[f"blocks.{l}.ln2.w"]

                if center_weights:
                    # Center the weights that read in from the LayerNormPre
                    state_dict[f"blocks.{l}.mlp.W_in"] -= einops.reduce(
                        state_dict[f"blocks.{l}.mlp.W_in"],
                        "d_model d_mlp -> 1 d_mlp",
                        "mean",
                    )

                if self.cfg.activation_name.startswith("solu"):
                    # Fold ln3 into activation
                    if fold_biases:
                        state_dict[f"blocks.{l}.mlp.b_out"] = state_dict[
                            f"blocks.{l}.mlp.b_out"
                        ] + (
                            state_dict[f"blocks.{l}.mlp.W_out"]
                            * state_dict[f"blocks.{l}.mlp.ln.b"][:, None]
                        ).sum(
                            -2
                        )

                        del state_dict[f"blocks.{l}.mlp.ln.b"]

                    state_dict[f"blocks.{l}.mlp.W_out"] = (
                        state_dict[f"blocks.{l}.mlp.W_out"]
                        * state_dict[f"blocks.{l}.mlp.ln.w"][:, None]
                    )

                    if center_weights:
                        # Center the weights that read in from the LayerNormPre
                        state_dict[f"blocks.{l}.mlp.W_out"] -= einops.reduce(
                            state_dict[f"blocks.{l}.mlp.W_out"],
                            "d_mlp d_model -> 1 d_model",
                            "mean",
                        )

                    del state_dict[f"blocks.{l}.mlp.ln.w"]

        if not self.cfg.final_rms and fold_biases:
            # Dumb bug from my old SoLU training code, some models have RMSNorm instead of LayerNorm
            # pre unembed.
            state_dict[f"head.b_H"] = state_dict[f"head.b_H"] + (
                state_dict[f"head.W_H"] * state_dict[f"ln_final.b"][:, None]
            ).sum(dim=-2)
            del state_dict[f"ln_final.b"]

        state_dict[f"head.W_H"] = (
            state_dict[f"head.W_H"] * state_dict[f"ln_final.w"][:, None]
        )
        del state_dict[f"ln_final.w"]

        if center_weights:
            # Center the weights that read in from the LayerNormPre
            state_dict[f"head.W_H"] -= einops.reduce(
                state_dict[f"head.W_H"], "d_model n_classes -> 1 n_classes", "mean"
            )
                    
        print("LayerNorm folded.")

        return state_dict
    
    def center_writing_weights(self, state_dict: Dict[str, torch.Tensor]):
        """Center Writing Weights.

        Centers the weights of the model that write to the residual stream - W_out, W_E, W_pos and
        W_out. This is done by subtracting the mean of the weights from the weights themselves. This
        is done in-place. See fold_layer_norm for more details.
        """
        # state_dict["embed.W_E"] = state_dict["embed.W_E"] - state_dict[
        #     "embed.W_E"
        # ].mean(-1, keepdim=True)
        if self.cfg.positional_embedding_type != "rotary":
            state_dict["pos_embed.W_pos"] = state_dict["pos_embed.W_pos"] - state_dict[
                "pos_embed.W_pos"
            ].mean(-1, keepdim=True)
        for l in range(self.cfg.n_layers):
            state_dict[f"blocks.{l}.attn.W_O"] = state_dict[
                f"blocks.{l}.attn.W_O"
            ] - state_dict[f"blocks.{l}.attn.W_O"].mean(
                -1, keepdim=True
            )  # W_O is [head_index, d_model, d_head]
            state_dict[f"blocks.{l}.attn.b_O"] = (
                state_dict[f"blocks.{l}.attn.b_O"]
                - state_dict[f"blocks.{l}.attn.b_O"].mean()
            )  # b_O is [d_model]
            if not self.cfg.attn_only:
                state_dict[f"blocks.{l}.mlp.W_out"] = state_dict[
                    f"blocks.{l}.mlp.W_out"
                ] - state_dict[f"blocks.{l}.mlp.W_out"].mean(-1, keepdim=True)
                state_dict[f"blocks.{l}.mlp.b_out"] = (
                    state_dict[f"blocks.{l}.mlp.b_out"]
                    - state_dict[f"blocks.{l}.mlp.b_out"].mean()
                )
                
        print("Centered weights writing to residual stream")
        return state_dict

    def fold_value_biases(self, state_dict: Dict[str, torch.Tensor]):
        """Fold the value biases into the output bias.

        Because attention patterns add up to 1, the value biases always have a constant effect on a
        head's output. Further, as the outputs of each head in a layer add together, each head's
        value bias has a constant effect on the *layer's* output, which can make it harder to
        interpret the effect of any given head, and it doesn't matter which head a bias is
        associated with. We can factor this all into a single output bias to the layer, and make it
        easier to interpret the head's output. Formally, we take b_O_new = b_O_original +
        sum_head(b_V_head @ W_O_head).
        """
        for layer in range(self.cfg.n_layers):
            # shape [head_index, d_head]
            if self.cfg.n_key_value_heads is None:
                b_V = state_dict[f"blocks.{layer}.attn.b_V"]
            else:
                b_V = state_dict[f"blocks.{layer}.attn._b_V"]
                b_V = torch.repeat_interleave(
                    b_V, dim=0, repeats=self.cfg.n_heads // self.cfg.n_key_value_heads
                )
            # [head_index, d_head, d_model]
            W_O = state_dict[f"blocks.{layer}.attn.W_O"]
            # [d_model]
            b_O_original = state_dict[f"blocks.{layer}.attn.b_O"]
            folded_b_O = b_O_original + (b_V[:, :, None] * W_O).sum([0, 1])

            state_dict[f"blocks.{layer}.attn.b_O"] = folded_b_O
            if self.cfg.n_key_value_heads is None:
                state_dict[f"blocks.{layer}.attn.b_V"] = torch.zeros_like(b_V)
            else:
                state_dict[f"blocks.{layer}.attn._b_V"] = torch.zeros_like(
                    state_dict[f"blocks.{layer}.attn._b_V"]
                )
                

        return state_dict

    def refactor_factored_attn_matrices(self, state_dict: Dict[str, torch.Tensor]):
        """Experimental method for managing queries, keys and values.

        As argued in [A Mathematical Framework for Transformer
        Circuits](https://transformer-circuits.pub/2021/framework/index.html), queries, keys and
        values are somewhat arbitrary intermediate terms when computing with the low rank factored
        matrices W_QK = W_Q @ W_K.T and W_OV = W_V @ W_O, and these matrices are the only thing
        determining head behaviour. But there are many ways to find a low rank factorization to a
        given matrix, and hopefully some of these are more interpretable than others! This method is
        one attempt, which makes all of the matrices have orthogonal rows or columns, W_O into a
        rotation and W_Q and W_K having the nth column in each having the same norm. The formula is
        $W_V = U @ S,W_O=Vh.T,W_Q=U@S.sqrt(),W_K=Vh@S.sqrt()$.

        More details:

        If W_OV = U @ S @ Vh.T in its singular value decomposition, (where S is in R^d_head not
        R^d_model, as W_OV is low rank), W_OV = (U @ S) @ (Vh.T) is an equivalent low rank
        factorisation, where rows/columns of each matrix are orthogonal! So setting $W_V=US$ and
        $W_O=Vh.T$ works just as well. I *think* this is a more interpretable setup, because now
        $W_O$ is just a rotation, and doesn't change the norm, so $z$ has the same norm as the
        result of the head.

        For $W_QK = W_Q @ W_K.T$ we use the refactor $W_Q = U @ S.sqrt()$ and $W_K = Vh @ S.sqrt()$,
        which is also equivalent ($S==S.sqrt() @ S.sqrt()$ as $S$ is diagonal). Here we keep the
        matrices as having the same norm, since there's not an obvious asymmetry between the keys
        and queries.

        Biases are more fiddly to deal with. For OV it's pretty easy - we just need (x @ W_V + b_V)
        @ W_O + b_O to be preserved, so we can set b_V' = 0. and b_O' = b_V @ W_O + b_O (note that
        b_V in R^{head_index x d_head} while b_O in R^{d_model}, so we need to sum b_V @ W_O along
        the head_index dimension too).

        For QK it's messy - we need to preserve the bilinear form of (x @ W_Q + b_Q) * (y @ W_K +
        b_K), which is fairly messy. To deal with the biases, we concatenate them to W_Q and W_K to
        simulate a d_model+1 dimensional input (whose final coordinate is always 1), do the SVD
        factorization on this effective matrix, then separate out into final weights and biases.
        """

        assert (
            self.cfg.positional_embedding_type != "rotary"
        ), "You can't refactor the QK circuit when using rotary embeddings (as the QK matrix depends on the position of the query and key)"

        for l in range(self.cfg.n_layers):
            # W_QK = W_Q @ W_K.T
            # Concatenate biases to make a d_model+1 input dimension
            W_Q_eff = torch.cat(
                [
                    state_dict[f"blocks.{l}.attn.W_Q"],
                    state_dict[f"blocks.{l}.attn.b_Q"][:, None, :],
                ],
                dim=1,
            )
            W_K_eff = torch.cat(
                [
                    state_dict[f"blocks.{l}.attn.W_K"],
                    state_dict[f"blocks.{l}.attn.b_K"][:, None, :],
                ],
                dim=1,
            )

            W_Q_eff_even, W_K_eff_even_T = (
                FactoredMatrix(W_Q_eff, W_K_eff.transpose(-1, -2)).make_even().pair
            )
            W_K_eff_even = W_K_eff_even_T.transpose(-1, -2)

            state_dict[f"blocks.{l}.attn.W_Q"] = W_Q_eff_even[:, :-1, :]
            state_dict[f"blocks.{l}.attn.b_Q"] = W_Q_eff_even[:, -1, :]
            state_dict[f"blocks.{l}.attn.W_K"] = W_K_eff_even[:, :-1, :]
            state_dict[f"blocks.{l}.attn.b_K"] = W_K_eff_even[:, -1, :]

            # W_OV = W_V @ W_O
            W_V = state_dict[f"blocks.{l}.attn.W_V"]
            W_O = state_dict[f"blocks.{l}.attn.W_O"]

            # Factors the bias to be consistent.
            b_V = state_dict[f"blocks.{l}.attn.b_V"]
            b_O = state_dict[f"blocks.{l}.attn.b_O"]
            effective_bias = b_O + einsum(
                "head_index d_head, head_index d_head d_model -> d_model", b_V, W_O
            )
            state_dict[f"blocks.{l}.attn.b_V"] = torch.zeros_like(b_V)
            state_dict[f"blocks.{l}.attn.b_O"] = effective_bias

            # Helper class to efficiently deal with low rank factored matrices.
            W_OV = FactoredMatrix(W_V, W_O)
            U, S, Vh = W_OV.svd()
            state_dict[f"blocks.{l}.attn.W_V"] = U @ S.diag_embed()
            state_dict[f"blocks.{l}.attn.W_O"] = transpose(Vh)

        return state_dict

    def load_and_process_state_dict(
        self,
        state_dict: Dict[str, torch.Tensor],
        fold_ln: Optional[bool] = True,
        center_writing_weights: Optional[bool] = True,
        fold_value_biases: Optional[bool] = True,
        refactor_factored_attn_matrices: Optional[bool] = False,
    ):
        """Load & Process State Dict.

        Load a state dict into the model, and to apply processing to simplify it. The state dict is
        assumed to be in the HookedTransformer format.

        See the relevant method (same name as the flag) for more details on the folding, centering
        and processing flags.

        Args:
            state_dict (dict): The state dict of the model, in HookedTransformer format. fold_ln
            fold_ln (bool, optional): Whether to fold in the LayerNorm weights to the
                subsequent linear layer. This does not change the computation. Defaults to True.
            center_writing_weights (bool, optional): Whether to center weights writing to the
                residual stream (ie set mean to be zero). Due to LayerNorm this doesn't change the
                computation. Defaults to True.
            fold_value_biases (bool, optional): Whether to fold the value biases into the output
                bias. Because attention patterns add up to 1, the value biases always have a
                constant effect on a layer's output, and it doesn't matter which head a bias is
                associated with. We can factor this all into a single output bias to the layer, and
                make it easier to interpret the head's output.
            refactor_factored_attn_matrices (bool, optional): Whether to convert the factored
                matrices (W_Q & W_K, and W_O & W_V) to be "even". Defaults to False.
            model_name (str, optional): checks the model name for special cases of state dict
                loading. Only used for Redwood 2L model currently.
        """
        if self.cfg.dtype not in [torch.float32, torch.float64] and fold_ln:
            logging.warning(
                "With reduced precision, it is advised to use `from_pretrained_no_processing` instead of `from_pretrained`."
            )

        state_dict = fill_missing_keys(self, state_dict)
        if fold_ln:
            if self.cfg.normalization_type in ["LN", "LNPre"]:
                state_dict = self.fold_layer_norm(state_dict)
            elif self.cfg.normalization_type in ["RMS", "RMSPre"]:
                state_dict = self.fold_layer_norm(
                    state_dict, fold_biases=False, center_weights=False
                )
            else:
                logging.warning(
                    "You are not using LayerNorm or RMSNorm, so the layer norm weights can't be folded! Skipping"
                )

        if center_writing_weights:
            if self.cfg.normalization_type not in ["LN", "LNPre"]:
                logging.warning(
                    "You are not using LayerNorm, so the writing weights can't be centered! Skipping"
                )
            elif self.cfg.final_rms:
                logging.warning(
                    "This model is using final RMS normalization, so the writing weights can't be centered! Skipping"
                )
            else:
                state_dict = self.center_writing_weights(state_dict)

        if fold_value_biases:
            state_dict = self.fold_value_biases(state_dict)
        if refactor_factored_attn_matrices:
            state_dict = self.refactor_factored_attn_matrices(state_dict)

        self.load_state_dict(state_dict, strict=False)

    def cuda(self):
        """Wrapper around cuda that also changes `self.cfg.device`."""
        return self.to("cuda")

    def cpu(self):
        """Wrapper around cuda that also changes `self.cfg.device`."""
        return self.to("cpu")

    def mps(self):
        """Wrapper around mps that also changes `self.cfg.device`."""
        return self.to("mps")

    def move_model_modules_to_device(self):
        self.embed.to(devices.get_device_for_block_index(0, self.cfg))
        self.hook_embed.to(devices.get_device_for_block_index(0, self.cfg))
        if self.cfg.positional_embedding_type != "rotary":
            self.pos_embed.to(devices.get_device_for_block_index(0, self.cfg))
            self.hook_pos_embed.to(devices.get_device_for_block_index(0, self.cfg))
        if hasattr(self, "ln_final"):
            self.ln_final.to(
                devices.get_device_for_block_index(self.cfg.n_layers - 1, self.cfg)
            )
        for i, block in enumerate(self.blocks):
            block.to(devices.get_device_for_block_index(i, self.cfg))

    @classmethod
    def from_pretrained(
        cls, 
        model_name: str,
        is_timm: bool = True,
        is_clip: bool = False,
        fold_ln: Optional[bool] = True,
        center_writing_weights: Optional[bool] = True,
        refactor_factored_attn_matrices: Optional[bool] = False,
        checkpoint_index: Optional[int] = None,
        checkpoint_value: Optional[int] = None,
        hf_model: Optional[ViTForImageClassification] = None,
        device: Optional[Union[str, torch.device]] = None,
        n_devices: Optional[int] = 1,
        move_to_device: Optional[bool] = True,
        fold_value_biases: Optional[bool] = True,
        default_prepend_bos: Optional[bool] = True,
        default_padding_side: Optional[Literal["left", "right"]] = "right",
        dtype="float32",
        use_attn_result: Optional[bool] = False,
        **from_pretrained_kwargs,
    ) -> "HookedViT":
        
        assert not (
            from_pretrained_kwargs.get("load_in_8bit", False)
            or from_pretrained_kwargs.get("load_in_4bit", False)
        ), "Quantization not supported"

        if isinstance(dtype, str):
            # Convert from string to a torch dtype
            dtype = DTYPE_FROM_STRING[dtype]
        if "torch_dtype" in from_pretrained_kwargs:
            # For backwards compatibility with the previous way to do low precision loading
            # This should maybe check the user did not explicitly set dtype *and* torch_dtype
            dtype = from_pretrained_kwargs["torch_dtype"]

        if (
            (from_pretrained_kwargs.get("torch_dtype", None) == torch.float16)
            or dtype == torch.float16
        ) and device in ["cpu", None]:
            logging.warning(
                "float16 models may not work on CPU. Consider using a GPU or bfloat16."
            )

        # Set up other parts of transformer
        
        cfg = convert_pretrained_model_config(
            model_name,
            is_timm=is_timm,
            is_clip=is_clip,
        )



        state_dict = get_pretrained_state_dict(
            model_name, is_timm, is_clip, cfg, hf_model, dtype=dtype, **from_pretrained_kwargs
        )

        model = cls(cfg, move_to_device=False)

        model.load_and_process_state_dict(
            state_dict,
            fold_ln=fold_ln,
            center_writing_weights=center_writing_weights,
            fold_value_biases=fold_value_biases,
            refactor_factored_attn_matrices=refactor_factored_attn_matrices,
        )

        # Set up other parameters
        model.set_use_attn_result(use_attn_result)


        if move_to_device:
            model.move_model_modules_to_device()

        print(f"Loaded pretrained model {model_name} into HookedTransformer")

        return model

    def set_use_attn_result(self, use_attn_result: bool):
        """Toggle whether to explicitly calculate and expose the result for each attention head.

        Useful for interpretability but can easily burn through GPU memory.
        """
        self.cfg.use_attn_result = use_attn_result

    def set_use_split_qkv_input(self, use_split_qkv_input: bool):
        """
        Toggles whether to allow editing of inputs to each attention head.
        """
        self.cfg.use_split_qkv_input = use_split_qkv_input

    def set_use_hook_mlp_in(self, use_hook_mlp_in: bool):
        """Toggles whether to allow storing and editing inputs to each MLP layer."""

        assert not self.cfg.attn_only, "Can't use hook_mlp_in with attn_only model"
        self.cfg.use_hook_mlp_in = use_hook_mlp_in

    def set_use_attn_in(self, use_attn_in: bool):
        """
        Toggles whether to allow editing of inputs to each attention head.
        """
        self.cfg.use_attn_in = use_attn_in

    def check_hooks_to_add(
        self,
        hook_point,
        hook_point_name,
        hook,
        dir="fwd",
        is_permanent=False,
        prepend=False,
    ) -> None:
        if hook_point_name.endswith("attn.hook_result"):
            assert (
                self.cfg.use_attn_result
            ), f"Cannot add hook {hook_point_name} if use_attn_result_hook is False"
        if hook_point_name.endswith(("hook_q_input", "hook_k_input", "hook_v_input")):
            assert (
                self.cfg.use_split_qkv_input
            ), f"Cannot add hook {hook_point_name} if use_split_qkv_input is False"
        if hook_point_name.endswith("mlp_in"):
            assert (
                self.cfg.use_hook_mlp_in
            ), f"Cannot add hook {hook_point_name} if use_hook_mlp_in is False"
        if hook_point_name.endswith("attn_in"):
            assert (
                self.cfg.use_attn_in
            ), f"Cannot add hook {hook_point_name} if use_attn_in is False"



"""
Prisma Repo
By Sonia Joseph

Copyright (c) Sonia Joseph. All rights reserved.

Inspired by TransformerLens. Some functions have been adapted from the TransformerLens project.
For more information on TransformerLens, visit: https://github.com/neelnanda-io/TransformerLens
"""

import logging

from transformers import AutoConfig, ViTForImageClassification, VivitForVideoClassification, CLIPModel

import timm
from vit_prisma.configs.HookedViTConfig import HookedViTConfig

import torch

from typing import Dict

import einops




def convert_clip_weights(
        old_state_dict,
        old_head_state_dict,
        cfg: HookedViTConfig,
):
    
    new_vision_model_state_dict = {}


    new_vision_model_state_dict["cls_token"] = old_state_dict["embeddings.class_embedding"].unsqueeze(0).unsqueeze(0)
    new_vision_model_state_dict["pos_embed.W_pos"] = old_state_dict["embeddings.position_embedding.weight"]
    new_vision_model_state_dict["embed.proj.weight"] = old_state_dict["embeddings.patch_embedding.weight"]
    new_vision_model_state_dict["embed.proj.bias"] =  torch.zeros((cfg.d_model,), device=new_vision_model_state_dict["embed.proj.weight"].device)
    new_vision_model_state_dict["ln_final.w"] = old_state_dict["post_layernorm.weight"]
    new_vision_model_state_dict["ln_final.b"] = old_state_dict["post_layernorm.bias"]
    new_vision_model_state_dict["ln_pre.w"] = old_state_dict["pre_layrnorm.weight"] #typo in ClipModel
    new_vision_model_state_dict["ln_pre.b"] = old_state_dict["pre_layrnorm.bias"]


    for layer in range(cfg.n_layers):
        layer_key = f"encoder.layers.{layer}"
        new_layer_key = f"blocks.{layer}"

        new_vision_model_state_dict[f"{new_layer_key}.ln1.w"] = old_state_dict[f"{layer_key}.layer_norm1.weight"]
        new_vision_model_state_dict[f"{new_layer_key}.ln1.b"] = old_state_dict[f"{layer_key}.layer_norm1.bias"]
        new_vision_model_state_dict[f"{new_layer_key}.ln2.w"] = old_state_dict[f"{layer_key}.layer_norm2.weight"]
        new_vision_model_state_dict[f"{new_layer_key}.ln2.b"] = old_state_dict[f"{layer_key}.layer_norm2.bias"]

        W_Q = old_state_dict[f"{layer_key}.self_attn.q_proj.weight"]
        W_K = old_state_dict[f"{layer_key}.self_attn.k_proj.weight"]
        W_V = old_state_dict[f"{layer_key}.self_attn.v_proj.weight"]
        W_O = old_state_dict[f"{layer_key}.self_attn.out_proj.weight"]

        W_Q = einops.rearrange(W_Q, "(h dh) d-> h d dh", h=cfg.n_heads, d=cfg.d_model, dh=cfg.d_head)
        W_K = einops.rearrange(W_K, "(h dh) d-> h d dh", h=cfg.n_heads, d=cfg.d_model, dh=cfg.d_head)
        W_V = einops.rearrange(W_V, "(h dh) d-> h d dh", h=cfg.n_heads, d=cfg.d_model, dh=cfg.d_head)
        W_O = einops.rearrange(W_O, "d (h dh) -> h dh d", h=cfg.n_heads, d=cfg.d_model, dh=cfg.d_head)
        
        new_vision_model_state_dict[f"{new_layer_key}.attn.W_Q"] = W_Q
        new_vision_model_state_dict[f"{new_layer_key}.attn.W_K"] = W_K
        new_vision_model_state_dict[f"{new_layer_key}.attn.W_V"] = W_V
        new_vision_model_state_dict[f"{new_layer_key}.attn.W_O"] = W_O

        b_Q = old_state_dict[f"{layer_key}.self_attn.q_proj.bias"]
        b_K = old_state_dict[f"{layer_key}.self_attn.k_proj.bias"]
        b_V = old_state_dict[f"{layer_key}.self_attn.v_proj.bias"]
        b_O = old_state_dict[f"{layer_key}.self_attn.out_proj.bias"]

        b_Q = einops.rearrange(b_Q, "(h dh) -> h dh", h=cfg.n_heads, dh=cfg.d_head)
        b_K = einops.rearrange(b_K, "(h dh) -> h dh", h=cfg.n_heads, dh=cfg.d_head)
        b_V = einops.rearrange(b_V, "(h dh) -> h dh", h=cfg.n_heads, dh=cfg.d_head)

        new_vision_model_state_dict[f"{new_layer_key}.attn.b_Q"] = b_Q
        new_vision_model_state_dict[f"{new_layer_key}.attn.b_K"] = b_K
        new_vision_model_state_dict[f"{new_layer_key}.attn.b_V"] = b_V
        new_vision_model_state_dict[f"{new_layer_key}.attn.b_O"] = b_O

        mlp_W_in = old_state_dict[f"{layer_key}.mlp.fc1.weight"]
        mlp_W_out = old_state_dict[f"{layer_key}.mlp.fc2.weight"]
        mlp_b_in = old_state_dict[f"{layer_key}.mlp.fc1.bias"]
        mlp_b_out = old_state_dict[f"{layer_key}.mlp.fc2.bias"]

        mlp_W_in = einops.rearrange(mlp_W_in, "m d -> d m")
        mlp_W_out = einops.rearrange(mlp_W_out, "d m -> m d")

        new_vision_model_state_dict[f"{new_layer_key}.mlp.W_in"] = mlp_W_in
        new_vision_model_state_dict[f"{new_layer_key}.mlp.W_out"] = mlp_W_out
        new_vision_model_state_dict[f"{new_layer_key}.mlp.b_in"] = mlp_b_in
        new_vision_model_state_dict[f"{new_layer_key}.mlp.b_out"] = mlp_b_out

    new_vision_model_state_dict["head.W_H"] = einops.rearrange(old_head_state_dict["weight"], "c d -> d c")
    new_vision_model_state_dict["head.b_H"] = torch.zeros((cfg.n_classes,), device=new_vision_model_state_dict["head.W_H"].device)

        
    return new_vision_model_state_dict


def convert_timm_weights(
        old_state_dict,
        cfg: HookedViTConfig,
):

    new_state_dict = {}
    new_state_dict["cls_token"] = old_state_dict["cls_token"]
    new_state_dict["pos_embed.W_pos"] = old_state_dict["pos_embed"].squeeze(0)
    new_state_dict["embed.proj.weight"] = old_state_dict["patch_embed.proj.weight"]
    new_state_dict["embed.proj.bias"] = old_state_dict["patch_embed.proj.bias"] 
    new_state_dict["ln_final.w"] = old_state_dict["norm.weight"]
    new_state_dict["ln_final.b"] = old_state_dict["norm.bias"]

    for layer in range(cfg.n_layers):
        layer_key = f"blocks.{layer}" 
        new_state_dict[f"{layer_key}.ln1.w"] = old_state_dict[f"{layer_key}.norm1.weight"]
        new_state_dict[f"{layer_key}.ln1.b"] = old_state_dict[f"{layer_key}.norm1.bias"]
        new_state_dict[f"{layer_key}.ln2.w"] = old_state_dict[f"{layer_key}.norm2.weight"]
        new_state_dict[f"{layer_key}.ln2.b"] = old_state_dict[f"{layer_key}.norm2.bias"]

        W = old_state_dict[f"{layer_key}.attn.qkv.weight"]
        W_reshape = einops.rearrange( W, "(three h dh) d ->three h d dh" , three=3, h=cfg.n_heads, d=cfg.d_model, dh=cfg.d_head)
        W_Q, W_K, W_V = torch.unbind(W_reshape, dim=0)
        new_state_dict[f"{layer_key}.attn.W_Q"] = W_Q
        new_state_dict[f"{layer_key}.attn.W_K"] = W_K
        new_state_dict[f"{layer_key}.attn.W_V"] = W_V

        W_O = old_state_dict[f"{layer_key}.attn.proj.weight"]
        W_O = einops.rearrange(W_O, "m (i h)->i h m", i=cfg.n_heads)
        new_state_dict[f"{layer_key}.attn.W_O"] = W_O

        attn_bias = old_state_dict[f"{layer_key}.attn.qkv.bias"]
        attn_bias_reshape = einops.rearrange(attn_bias, "(three h dh) -> three h dh", three=3, h=cfg.n_heads, dh=cfg.d_head)
        b_Q, b_K, b_V = torch.unbind(attn_bias_reshape, dim=0)
        new_state_dict[f"{layer_key}.attn.b_Q"] = b_Q
        new_state_dict[f"{layer_key}.attn.b_K"] = b_K
        new_state_dict[f"{layer_key}.attn.b_V"] = b_V

        b_O = old_state_dict[f"{layer_key}.attn.proj.bias"]
        new_state_dict[f"{layer_key}.attn.b_O"] = b_O

        new_state_dict[f"{layer_key}.mlp.b_in"] = old_state_dict[f"{layer_key}.mlp.fc1.bias"]
        new_state_dict[f"{layer_key}.mlp.b_out"] = old_state_dict[f"{layer_key}.mlp.fc2.bias"]

        mlp_W_in = old_state_dict[f"{layer_key}.mlp.fc1.weight"]
        mlp_W_in = einops.rearrange(mlp_W_in, "m d -> d m")
        new_state_dict[f"{layer_key}.mlp.W_in"] = mlp_W_in

        mlp_W_out = old_state_dict[f"{layer_key}.mlp.fc2.weight"]
        mlp_W_out = einops.rearrange(mlp_W_out, "d m -> m d")
        new_state_dict[f"{layer_key}.mlp.W_out"] = mlp_W_out



    new_state_dict["head.W_H"] = einops.rearrange(old_state_dict["head.weight"], "c d -> d c")
    new_state_dict["head.b_H"] = old_state_dict["head.bias"]


    return new_state_dict


def convert_vivet_weights(
        old_state_dict,
        cfg: HookedViTConfig,
):

    new_state_dict = {}

    new_state_dict["cls_token"] = old_state_dict["vivit.embeddings.cls_token"]
    new_state_dict["pos_embed.W_pos"] = old_state_dict["vivit.embeddings.position_embeddings"].squeeze(0)
    new_state_dict["embed.proj.weight"] = old_state_dict["vivit.embeddings.patch_embeddings.projection.weight"]
    new_state_dict["embed.proj.bias"] = old_state_dict["vivit.embeddings.patch_embeddings.projection.bias"] 
    new_state_dict["ln_final.w"] = old_state_dict["vivit.layernorm.weight"]
    new_state_dict["ln_final.b"] = old_state_dict["vivit.layernorm.bias"]

    for layer in range(cfg.n_layers):
        layer_key = f"vivit.encoder.layer.{layer}" 
        new_layer_key = f"blocks.{layer}"
        new_state_dict[f"{new_layer_key}.ln1.w"] = old_state_dict[f"{layer_key}.layernorm_before.weight"]
        new_state_dict[f"{new_layer_key}.ln1.b"] = old_state_dict[f"{layer_key}.layernorm_before.bias"]
        new_state_dict[f"{new_layer_key}.ln2.w"] = old_state_dict[f"{layer_key}.layernorm_after.weight"]
        new_state_dict[f"{new_layer_key}.ln2.b"] = old_state_dict[f"{layer_key}.layernorm_after.bias"]

        W_Q  = old_state_dict[f"{layer_key}.attention.attention.query.weight"]
        W_K  = old_state_dict[f"{layer_key}.attention.attention.key.weight"]
        W_V  = old_state_dict[f"{layer_key}.attention.attention.value.weight"]

        new_state_dict[f"{new_layer_key}.attn.W_Q"] = einops.rearrange(W_Q, "(h dh) d -> h d dh", h=cfg.n_heads, d=cfg.d_model, dh=cfg.d_head)
        new_state_dict[f"{new_layer_key}.attn.W_K"] = einops.rearrange(W_K, "(h dh) d -> h d dh", h=cfg.n_heads, d=cfg.d_model, dh=cfg.d_head)
        new_state_dict[f"{new_layer_key}.attn.W_V"] = einops.rearrange(W_V, "(h dh) d -> h d dh", h=cfg.n_heads, d=cfg.d_model, dh=cfg.d_head)

        W_O = old_state_dict[f"{layer_key}.attention.output.dense.weight"]
        new_state_dict[f"{new_layer_key}.attn.W_O"] =  einops.rearrange(W_O, "d (h dh) -> h dh d", h=cfg.n_heads, d=cfg.d_model, dh=cfg.d_head)


        b_Q  = old_state_dict[f"{layer_key}.attention.attention.query.bias"]
        b_K  = old_state_dict[f"{layer_key}.attention.attention.key.bias"]
        b_V  = old_state_dict[f"{layer_key}.attention.attention.value.bias"]

        new_state_dict[f"{new_layer_key}.attn.b_Q"] = einops.rearrange(b_Q, "(h dh) -> h dh",  h=cfg.n_heads, dh=cfg.d_head)
        new_state_dict[f"{new_layer_key}.attn.b_K"] = einops.rearrange(b_K, "(h dh) -> h dh",  h=cfg.n_heads, dh=cfg.d_head)
        new_state_dict[f"{new_layer_key}.attn.b_V"] = einops.rearrange(b_V, "(h dh) -> h dh",  h=cfg.n_heads, dh=cfg.d_head)

        b_O = old_state_dict[f"{layer_key}.attention.output.dense.bias"]
        new_state_dict[f"{new_layer_key}.attn.b_O"] = b_O

        mlp_W_in = old_state_dict[f"{layer_key}.intermediate.dense.weight"]
        new_state_dict[f"{new_layer_key}.mlp.W_in"] =  einops.rearrange(mlp_W_in, "m d -> d m")

        mlp_W_out  = old_state_dict[f"{layer_key}.output.dense.weight"]
       
        new_state_dict[f"{new_layer_key}.mlp.W_out"] = einops.rearrange(mlp_W_out, "d m -> m d")

        new_state_dict[f"{new_layer_key}.mlp.b_in"] =  old_state_dict[f"{layer_key}.intermediate.dense.bias"]
        new_state_dict[f"{new_layer_key}.mlp.b_out"] =  old_state_dict[f"{layer_key}.output.dense.bias"]


    new_state_dict["head.W_H"] = einops.rearrange(old_state_dict["classifier.weight"], "c d -> d c")
    new_state_dict["head.b_H"] = old_state_dict["classifier.bias"]


    return new_state_dict

def get_pretrained_state_dict(
    official_model_name: str,
    is_timm: bool,
    is_clip: bool,
    cfg: HookedViTConfig,
    hf_model=None,
    dtype: torch.dtype = torch.float32,
    **kwargs,
) -> Dict[str, torch.Tensor]:
    """
    Loads in the model weights for a pretrained model, and processes them to
    have the HookedTransformer parameter names and shapes. Supports checkpointed
    models (and expects the checkpoint info to be stored in the config object)

    hf_model: Optionally, a HuggingFace model object. If provided, we will use
        these weights rather than reloading the model.
    dtype: The dtype to load the HuggingFace model in.
    kwargs: Other optional arguments passed to HuggingFace's from_pretrained.
        Also given to other HuggingFace functions when compatible.
    """
    if "torch_dtype" in kwargs:
        dtype = kwargs["torch_dtype"]
        del kwargs["torch_dtype"]
    # official_model_name = get_official_model_name(official_model_name)
    # if official_model_name.startswith(NEED_REMOTE_CODE_MODELS) and not kwargs.get(
    #     "trust_remote_code", False
    # ):
    #     logging.warning(
    #         f"Loading model {official_model_name} state dict requires setting trust_remote_code=True"
    #     )
    #     kwargs["trust_remote_code"] = True
        
    try:
        if is_timm:
            hf_model = hf_model if hf_model is not None else timm.create_model(official_model_name, pretrained=True)
            for param in hf_model.parameters():
                param.requires_grad = False
            state_dict = convert_timm_weights(hf_model.state_dict(), cfg)
        elif is_clip:
            full_model = hf_model if hf_model is not None else CLIPModel.from_pretrained(official_model_name)
            for param in full_model.parameters():
                param.requires_grad = False
            vision = full_model.vision_model
            visual_projection = full_model.visual_projection
            state_dict = convert_clip_weights(vision.state_dict(), visual_projection.state_dict(), cfg)

        elif cfg.is_video_transformer:
            if "vivit" in official_model_name:
                hf_model = hf_model if hf_model is not None else VivitForVideoClassification.from_pretrained(official_model_name, torch_dtype=dtype, **kwargs)

                for param in hf_model.parameters():
                    param.requires_grad = False

                state_dict = convert_vivet_weights(hf_model.state_dict(), cfg)
            else:
                raise ValueError
        
        else:
            hf_model = hf_model if hf_model is not None else ViTForImageClassification.from_pretrained(
                    official_model_name, torch_dtype=dtype, **kwargs
            )
            raise ValueError

            # Load model weights, and fold in layer norm weights


                
        return state_dict

    except:
        raise ValueError(
            f"Loading weights from the architecture is not currently supported: {cfg.original_architecture}, generated from model name {cfg.model_name}. Feel free to open an issue on GitHub to request this feature."
        )

def fill_missing_keys(model, state_dict):
    """Takes in a state dict from a pretrained model, and fills in any missing keys with the default initialization.

    This function is assumed to be run before weights are initialized.

    Args:
        state_dict (dict): State dict from a pretrained model

    Returns:
        dict: State dict with missing keys filled in
    """
    # Get the default state dict
    default_state_dict = model.state_dict()
    # Get the keys that are missing from the pretrained model
    missing_keys = set(default_state_dict.keys()) - set(state_dict.keys())
    # Fill in the missing keys with the default initialization
    for key in missing_keys:
        if "hf_model" in key:
            # Skip keys that are from the HuggingFace model, if loading from HF.
            continue
        if "W_" in key:
            logging.warning(
                "Missing key for a weight matrix in pretrained, filled in with an empty tensor: {}".format(
                    key
                )
            )
        state_dict[key] = default_state_dict[key]
    return state_dict

def convert_pretrained_model_config(model_name: str, is_timm: bool = True, is_clip: bool = False) -> HookedViTConfig:
    
    

    if is_timm:
        model = timm.create_model(model_name)
        hf_config = AutoConfig.from_pretrained(model.default_cfg['hf_hub_id'])
    elif is_clip: # Extract vision encoder from dual-encoder CLIP model.
        hf_config = AutoConfig.from_pretrained(model_name).vision_config
        hf_config.architecture = 'vit_clip_vision_encoder'
        hf_config.num_classes = hf_config.projection_dim # final output dimension instead of classes
    else:
        hf_config = AutoConfig.from_pretrained(model_name)
        
#     print('hf config', hf_config)
            
 

    if hasattr(hf_config, 'patch_size'):
        ps = hf_config.patch_size
    elif hasattr(hf_config, "tubelet_size"):
        ps = hf_config.tubelet_size[1]

    pretrained_config = {
                    'n_layers' : hf_config.num_hidden_layers,
                    'd_model' : hf_config.hidden_size,
                    'd_head' : hf_config.hidden_size // hf_config.num_attention_heads,
                    'model_name' : hf_config._name_or_path,
                    'n_heads' : hf_config.num_attention_heads,
                    'd_mlp' : hf_config.intermediate_size,
                    'activation_name' : hf_config.hidden_act,
                    'eps' : hf_config.layer_norm_eps,
                    'original_architecture' : getattr(hf_config, 'architecture', getattr(hf_config, 'architectures', None)),
                    'initializer_range' : hf_config.initializer_range,
                    'n_channels' : hf_config.num_channels,
                    'patch_size' : ps,
                    'image_size' : hf_config.image_size,
                    'n_classes' : getattr(hf_config, "num_classes", None),
                    'n_params' : sum(p.numel() for p in model.parameters() if p.requires_grad) if is_timm else None,
                }
    
    # Rectifying Huggingface bugs:
    # Currently a bug getting configs, only this model confirmed to work and even it requires modification of eps
    if is_timm and model_name == "vit_base_patch16_224":
        pretrained_config.update({
            "eps": 1e-6,
            "return_type": "class_logits",
        })
    
    # Config for 32 is incorrect, fix manually 
    if is_timm and model_name == "vit_base_patch32_224":
        pretrained_config.update({
            "patch_size": 32,
            "eps": 1e-6,
            "return_type": "class_logits"
        })

    if is_clip:
        pretrained_config.update({
            "layer_norm_pre": True,
            "return_type": "class_logits" # actually returns 'visual_projection'
        })


    # Config is for ViVet, need to add more properties
    if hasattr(hf_config, "tubelet_size"):
        pretrained_config.update({
            "is_video_transformer": True,
            "video_tubelet_depth": hf_config.tubelet_size[0],
            "video_num_frames": hf_config.video_size[0],
            "n_classes": 400 if "kinetics400" in model_name else None,
            "return_type": "class_logits" if "kinetics400" in model_name else "pre_logits",

        })

    
    print(pretrained_config)

    return HookedViTConfig.from_dict(pretrained_config)



import torch.nn as nn
import torch

from typing import Dict, Optional, Tuple, Union

from jaxtyping import Float, Int

from vit_prisma.configs.HookedViTConfig import HookedViTConfig
from vit_prisma.prisma_tools.hook_point import HookPoint

class LayerNormPre(nn.Module):
    def __init__(self, cfg: Union[Dict, HookedViTConfig]):
        """LayerNormPre - the 'center and normalise' part of LayerNorm. Length is
        normally d_model, but is d_mlp for softmax. Not needed as a parameter. This
        should only be used in inference mode after folding in LayerNorm weights"""
        super().__init__()
        if isinstance(cfg, Dict):
            cfg = HookedViTConfig.from_dict(cfg)
        self.cfg = cfg
        self.eps = self.cfg.eps

        # Adds a hook point for the normalisation scale factor
        self.hook_scale = HookPoint()  # [batch, pos]
        # Hook Normalized captures LN output - here it's a vector with std 1 and mean 0
        self.hook_normalized = HookPoint()  # [batch, pos, length]

    def forward(
        self,
        x: Union[
            Float[torch.Tensor, "batch pos d_model"],
            Float[torch.Tensor, "batch pos head_index d_model"],
        ],
    ) -> Union[
        Float[torch.Tensor, "batch pos d_model"],
        Float[torch.Tensor, "batch pos head_index d_model"],
    ]:
        if self.cfg.dtype not in [torch.float32, torch.float64]:
            x = x.to(torch.float32)

        x = x - x.mean(axis=-1, keepdim=True)  # [batch, pos, length]
        scale: Union[
            Float[torch.Tensor, "batch pos 1"],
            Float[torch.Tensor, "batch pos head_index 1"],
        ] = self.hook_scale((x.pow(2).mean(-1, keepdim=True) + self.eps).sqrt())
        return self.hook_normalized(x / scale).to(self.cfg.dtype)


class LayerNorm(nn.Module):
    def __init__(
        self, cfg: Union[Dict, HookedViTConfig], length: Optional[int] = None
    ):
        """
        LayerNorm with optional length parameter

        length (Optional[int]): If the dimension of the LayerNorm. If not provided, assumed to be d_model
        """
        super().__init__()
        if isinstance(cfg, Dict):
            cfg = HookedViTConfig.from_dict(cfg)
        self.cfg = cfg
        self.eps = self.cfg.eps
        if length is None:
            self.length = self.cfg.d_model
        else:
            self.length = length

        self.w = nn.Parameter(torch.ones(self.length, dtype=cfg.dtype))
        self.b = nn.Parameter(torch.zeros(self.length, dtype=cfg.dtype))

        # Adds a hook point for the normalisation scale factor
        self.hook_scale = HookPoint()  # [batch, pos, 1]
        # Hook_normalized is on the LN output
        self.hook_normalized = HookPoint()  # [batch, pos, length]

    def forward(
        self,
        x: Union[
            Float[torch.Tensor, "batch pos d_model"],
            Float[torch.Tensor, "batch pos head_index d_model"],
        ],
    ) -> Union[
        Float[torch.Tensor, "batch pos d_model"],
        Float[torch.Tensor, "batch pos head_index d_model"],
    ]:
        if self.cfg.dtype not in [torch.float32, torch.float64]:
            x = x.to(torch.float32)

        x = x - x.mean(axis=-1, keepdim=True)  # [batch, pos, length]
        scale: Float[torch.Tensor, "batch pos 1"] = self.hook_scale(
            (x.pow(2).mean(-1, keepdim=True) + self.eps).sqrt()
        )
        x = x / scale  # [batch, pos, length]
        return self.hook_normalized(x * self.w + self.b).to(self.cfg.dtype)
 


import pytest
import torch

from vit_prisma.models.base_vit import HookedViT
from vit_prisma.configs.HookedViTConfig import HookedViTConfig

#Test taken from transformerlens with minor modifications

batch_size = 2
channels = 3
height = 224
width = 224
input_images = torch.rand((batch_size, channels, height, width))

# instantiate a untrained network
n_layers = 1
d_head = 8
d_model = 8
d_mlp=8

model = HookedViT(HookedViTConfig(n_layers,d_head,d_model, d_mlp, return_type="logits", activation_name="solu_ln"))

act_names_in_cache = [
    "hook_embed",
    "hook_pos_embed",
    "blocks.0.hook_resid_pre",
    "blocks.0.ln1.hook_scale",
    "blocks.0.ln1.hook_normalized",
    "blocks.0.attn.hook_q",
    "blocks.0.attn.hook_k",
    "blocks.0.attn.hook_v",
    "blocks.0.attn.hook_attn_scores",
    "blocks.0.attn.hook_pattern",
    "blocks.0.attn.hook_z",
    "blocks.0.hook_attn_out",
    "blocks.0.hook_resid_mid",
    "blocks.0.ln2.hook_scale",
    "blocks.0.ln2.hook_normalized",
    "blocks.0.mlp.hook_pre",
    "blocks.0.mlp.hook_mid",
    "blocks.0.mlp.ln.hook_scale",
    "blocks.0.mlp.ln.hook_normalized",
    "blocks.0.mlp.hook_post",
    "blocks.0.hook_mlp_out",
    "blocks.0.hook_resid_post",
    "ln_final.hook_scale",
    "ln_final.hook_normalized",
]


def test_cache_hook_names():
    _, cache = model.run_with_cache(input_images)
    assert list(cache.keys()) == act_names_in_cache


import json
import torch
# Save config along with trainig files


def object_to_dict(obj):
    if isinstance(obj, (int, float, str, bool, type(None))):
        return obj
    elif isinstance(obj, dict):
        return {k: object_to_dict(v) for k, v in obj.items() if not callable(v)}
    elif isinstance(obj, list):
        return [object_to_dict(item) for item in obj]
    elif hasattr(obj, "__dict__"):  # for custom objects
        # Filter out methods or built-in functions
        return object_to_dict({k: v for k, v in obj.__dict__.items() if not callable(v)})
    elif isinstance(obj, torch.dtype):
        return str(obj)
    else:
        raise TypeError(f"Object of type {type(obj)} is not serializable to JSON")

# Function to save config to a JSON file
def save_config_to_file(config, file_path):
    with open(file_path, 'w') as f:
        json.dump(object_to_dict(config), f, indent=4)




import numpy as np
import plotly.graph_objects as go
import plotly.express as px

import torch

from vit_prisma.utils.data_utils.imagenet_emoji import IMAGENET_EMOJI

def display_grid_on_image_with_heatmap(image, patch_dictionary, patch_size=32,
                                       layer_idx=-1,
                                       imagenet_class_to_emoji=IMAGENET_EMOJI,
                                       emoji_font_size=30,
                                       heatmap_mode='logit_values',
                                       alpha_color=.6,
                                       return_graph=False):

    valid_heatmap_modes = ['logit_values', 'emoji_colors']
    if heatmap_mode not in valid_heatmap_modes:
        raise ValueError(f"Invalid heatmap_mode '{heatmap_mode}'. Valid options are {valid_heatmap_modes}.")

    if isinstance(image, np.ndarray) and image.shape[-1] == 3:
        # Image is in correct format.
        pass
    elif isinstance(image, torch.Tensor):
        # Convert torch.Tensor to numpy.ndarray if necessary.
        image = image.detach().cpu().numpy().transpose(1, 2, 0)
        if image.max() <= 1.0:
            image *= 255  # Convert from [0, 1] to [0, 255] if needed.
        image = image.astype(np.uint8)

    grid_size_x = image.shape[1] // patch_size
    grid_size_y = image.shape[0] // patch_size

    # Initialize matrices with NaNs for logit values or zeros for emoji colors
    if heatmap_mode == 'logit_values':
        heatmap_matrix = np.full((grid_size_y, grid_size_x), np.nan)
    else:  # 'emoji_colors'
        heatmap_matrix = np.zeros((grid_size_y, grid_size_x))

    if imagenet_class_to_emoji is not None and heatmap_mode == 'emoji_colors':
        # Map emojis to unique integers if in 'emoji_colors' mode
        unique_emojis = {emoji: idx for idx, emoji in enumerate(set(imagenet_class_to_emoji.values()))}
    else:
        unique_emojis = {}

    fig = px.imshow(image)
    fig.update_traces(hoverinfo='none', hovertemplate=None)

    annotations = []
    hover_texts = []
    x_centers = []
    y_centers = []
    for patch_index, patch_data in sorted(patch_dictionary.items()):
        if patch_index == 0:
            continue  # Skip the CLS token.
        logit_data = patch_data[layer_idx]
        length_of_patch_data = len(logit_data)
        if length_of_patch_data >= 3:
            logit_value, class_name, class_index = logit_data[:3]
        else:
            print("Error in length of patch data.")
            continue

        adjusted_index = patch_index - 1
        row, col = divmod(adjusted_index, grid_size_x)

        if heatmap_mode == 'logit_values':
            heatmap_matrix[row, col] = logit_value
        else:  # 'emoji_colors'
            emoji = imagenet_class_to_emoji.get(class_index, "❓")
            heatmap_matrix[row, col] = unique_emojis.get(emoji, 0)

        emoji = imagenet_class_to_emoji.get(class_index, "❓") if imagenet_class_to_emoji is not None else "❓"
        annotations.append(go.layout.Annotation(x=col * patch_size + patch_size / 2,
                                                y=row * patch_size + patch_size / 2,
                                                text=emoji,
                                                showarrow=False,
                                                font=dict(size=emoji_font_size)))
        hover_texts.append(f"Patch: {patch_index}<br>Class: {class_name}<br>Logit: {logit_value:.2f}")
        x_centers.append(col * patch_size + patch_size / 2)
        y_centers.append(row * patch_size + patch_size / 2)

    # Choose color scale based on heatmap_mode
    colorscale = 'Viridis' if heatmap_mode == 'logit_values' else px.colors.qualitative.Plotly
    heatmap = go.Heatmap(z=heatmap_matrix, x0=0.5 * patch_size, dx=patch_size, y0=0.5 * patch_size, dy=patch_size,
                         colorscale=colorscale, opacity=alpha_color, showscale=(heatmap_mode == 'logit_values'), hoverinfo="none")
    fig.add_trace(heatmap)

    fig.add_trace(go.Scatter(x=x_centers, y=y_centers, mode='markers', marker=dict(color='rgba(0,0,0,0)'),
                             hoverinfo='text', text=hover_texts, hoverlabel=dict(bgcolor="white", font_size=12, font_family="Arial")))
    fig.update_layout(annotations=annotations)
    fig.update_layout(showlegend=False, plot_bgcolor='rgba(0,0,0,0)')
    fig.update_xaxes(showticklabels=False).update_yaxes(showticklabels=False)


    if return_graph:
      return fig
    else:
      fig.show()


# Animal logit lens

def display_patch_logit_lens(patch_dictionary, width=1000, height=1200, emoji_size=26, return_graph=False, show_colorbar=True, labels=None):
    num_patches = len(patch_dictionary)

    # Assuming data_array_formatted is correctly shaped according to your data structure
    data_array_formatted = np.array([[item[0] for item in list(patch_dictionary.values())[i]] for i in range(num_patches)])

    # Modify hover text generation based on whether labels are provided
    if labels:
      hover_text = [
            [f"{labels[j]}: {item[1]}" for j, item in enumerate(list(patch_dictionary.values())[i])]
            for i in range(num_patches)
        ] 
    else:
        hover_text = [[str(item[1]) for item in list(patch_dictionary.values())[i]] for i in range(num_patches)]

    # Creating the interactive heatmap
    fig = go.Figure(data=go.Heatmap(
        z=data_array_formatted,
        x=list(patch_dictionary.keys())[:num_patches],
        y=[f'{i}' for i in range(data_array_formatted.shape[0])],  # Patch Number
        hoverongaps=False,
        colorbar=dict(title='Logit Value') if show_colorbar else None,
        text=hover_text,
        hoverinfo="text"
    ))

    # Initialize a list to hold annotations for emojis
    annotations = []

    # Calculate half the distance between cells in both x and y directions for annotation placement
    x_half_dist = 0.5
    y_half_dist = 0.2

    for i, patch in enumerate(patch_dictionary.values()):
        for j, items in enumerate(patch):  # Extract class index directly from the patch_dictionary
            class_index = items[2]
            emoji = IMAGENET_EMOJI.get(class_index, "")  # Use class index for emoji lookup, default to empty if not found
            if emoji:  # Add annotation if emoji is found
                annotations.append(go.layout.Annotation(x=j + x_half_dist, y=i + y_half_dist, text=emoji, showarrow=False, font=dict(color="white", size=emoji_size)))

    # Add annotations to the figure
    fig.update_layout(annotations=annotations)

    # Configure the layout of the figure
    fig.update_layout(
        title='Per-Patch Logit Lens',
        xaxis=dict(title='Layer Number'),
        yaxis=dict(title='Patch Number'),
        autosize=False,
        width=width,
        height=height
    )
    fig.show()

    if return_graph:
        return fig


import torch.nn as nn
from vit_prisma.prisma_tools.hook_point import HookPoint
from vit_prisma.configs.HookedViTConfig import HookedViTConfig

from typing import Dict, Optional, Tuple, Union

from jaxtyping import Float, Int

import torch
import torch.nn.functional as F

import numpy as np

from vit_prisma.models.activation_fns import gelu_fast, gelu_new, solu

from vit_prisma.models.layers.layer_norm import LayerNorm, LayerNormPre

from fancy_einsum import einsum

class MLP(nn.Module):

    def __init__(self, cfg: Union[Dict, HookedViTConfig]):
        super().__init__()
        
        if not isinstance(cfg, HookedViTConfig):
            cfg = HookedViTConfig.from_dict(cfg)
        
        self.cfg = cfg

        self.W_in = nn.Parameter(
            torch.empty(self.cfg.d_model, self.cfg.d_mlp, dtype=self.cfg.dtype)
        )
        self.b_in = nn.Parameter(
            torch.empty(self.cfg.d_mlp, dtype=self.cfg.dtype)
        )
        self.W_out = nn.Parameter(
            torch.empty(self.cfg.d_mlp, self.cfg.d_model, dtype=self.cfg.dtype)
        )
        self.b_out = nn.Parameter(
            torch.empty(self.cfg.d_model, dtype=self.cfg.dtype)
        )

        self.hook_pre = HookPoint()
        self.hook_post = HookPoint()

        if self.cfg.activation_name == "relu":
                self.act_fn = F.relu
        elif self.cfg.activation_name == "gelu":
            self.act_fn = F.gelu
        elif self.cfg.activation_name == "silu":
            self.act_fn = F.silu
        elif self.cfg.activation_name == "gelu_new":
            self.act_fn = gelu_new
        elif self.cfg.activation_name == "gelu_fast":
            self.act_fn = gelu_fast
        elif self.cfg.activation_name == "solu_ln": # why does only solu have a layernorm? 
            self.act_fn = solu
            # Hook taken between activation and layer norm
            self.hook_mid = HookPoint()  # [batch, pos, d_mlp]
            if self.cfg.normalization_type == "LN":
                self.ln = LayerNorm(self.cfg, self.cfg.d_mlp)
            else:
                self.ln = LayerNormPre(self.cfg)

        else:
            raise ValueError(f"Invalid activation function name: {self.cfg.activation_name}")
    
    def forward(self, x: Float[torch.Tensor, "batch pos d_model"]
    ) -> Float[torch.Tensor, "batch pos d_model"]:
        
        pre_act = self.hook_pre(
              einsum("batch pos d_model, d_model d_mlp -> batch pos d_mlp", x, self.W_in)
            + self.b_in
        )
        if not self.cfg.activation_name.endswith("_ln"):
              post_act = self.hook_post(self.act_fn(pre_act))
        else:
              mid_act = self.hook_mid(self.act_fn(pre_act))
              post_act = self.hook_post(self.ln(mid_act))
        return (
              einsum("batch pos d_mlp, d_mlp d_model -> batch pos d_model", post_act, self.W_out) 
              + self.b_out
        )
          



# import unittest
# from torchvision import datasets, transforms
# from torch.nn import CrossEntropyLoss
# from vit_prisma.models.base_vit import BaseViT  # Assuming you have a ViT model defined somewhere

# from vit_prisma.configs.CircleConfig import GlobalConfig
# from vit_prisma.training.trainer import train
# from vit_prisma.dataloaders.induction import InductionDataset
# import timm
# import torch.nn as nn

# from vit_prisma.dataloaders.circle import CircleDataset



# class TestTrainingFunction(unittest.TestCase):

#         def setUp(self):
#             transform = transforms.Compose([
#                 transforms.ToPILImage('L'),  # Convert tensor to PIL Image. Ensure your data is suitable for this.
#                 transforms.Grayscale(num_output_channels=3),  # Convert grayscale to RGB
#                 transforms.Resize((224, 224)),  # Resize to the desired dimensions
#                 transforms.ToTensor(),  # Convert back to tensor
#                 transforms.Normalize(mean=[0.5], std=[0.5])  # Normalize the image
#             ])


#             self.train_dataset, self.val_dataset = CircleDataset('train', transform=transform), CircleDataset('test', transform=transform)
    
#             self.config = GlobalConfig()

#             # Define a simple model
#             self.model = timm.create_model('vit_base_patch16_224', pretrained=True)
#             self.model.head = nn.Linear(self.model.head.in_features, 60)
        
#             # Freeze all the parameters
#             for param in self.model.parameters():
#                 param.requires_grad = False
            
#             # Unfreeze the parameters of the classification head
#             for param in self.model.head.parameters():
#                 param.requires_grad = True

#         def test_train_function(self):
#             trained_model = train(self.model, self.config, self.train_dataset, self.val_dataset)
#             self.assertIsInstance(trained_model, type(self.model))

# if __name__ == '__main__':
#     unittest.main()



from dataclasses import dataclass
import torch.nn as nn
import torch

from typing import Any, Dict, List, Optional


@dataclass
class HookedViTConfig:

    n_layers: int = None
    d_model: int = None
    d_head: int = None
    d_mlp: int = None
    model_name: str = "custom"
    n_heads: int = 4
    activation_name: str = "gelu"
    d_vocab: int = -1
    eps: float = 1e-6
    use_attn_result: bool = False
    use_attn_scale: bool = True
    use_split_qkv_input: bool = False
    use_hook_mlp_in: bool = False
    use_attn_in: bool = False
    use_local_attn: bool = False
    original_architecture: Optional[str] = None
    from_checkpoint: bool = False
    checkpoint_index: Optional[int] = None
    checkpoint_label_type: Optional[str] = None
    checkpoint_value: Optional[int] = None
    tokenizer_name: Optional[str] = None
    window_size: Optional[int] = None
    attn_types: Optional[List] = None
    init_mode: str = "gpt2"
    normalization_type: Optional[str] = "LN"
    device: Optional[str] = 'cpu'
    n_devices: int = 1
    attention_dir: str = "bidirectional"
    attn_only: bool = False
    seed: Optional[int] = None
    initializer_range: float = -1.0
    init_weights: bool = True
    scale_attn_by_inverse_layer_idx: bool = False
    positional_embedding_type: str = "standard"
    final_rms: bool = False
    d_vocab_out: int = -1
    parallel_attn_mlp: bool = False
    rotary_dim: Optional[int] = None
    n_params: Optional[int] = None
    use_hook_tokens: bool = False
    gated_mlp: bool = False
    default_prepend_bos: bool = True
    dtype: torch.dtype = torch.float32
    tokenizer_prepends_bos: Optional[bool] = None
    n_key_value_heads: Optional[int] = None
    post_embedding_ln: bool = False
    rotary_base: int = 10000
    trust_remote_code: bool = False
    rotary_adjacent_pairs: bool = False
        
    # Layer norm
    layer_norm_pre: bool = False # add layernorm before transformer blocks

    #Bert Block
    use_bert_block: bool = False 

    # Initialization
    weight_type: str = 'he'
    cls_std: float = 1e-6
    pos_std: float = 0.02

    # Image related
    n_channels: int = 3
    patch_size: int = 32
    image_size: int = 224

    # Classification related
    classification_type: str = 'cls'
    n_classes: int = 10
    return_type: str = 'pre_logits'

    # Logging related
    log_dir: str = 'logs'
    use_wandb: bool = True
    wandb_team_name: str = 'perceptual-alignment'
    wandb_project_name: str = None
    log_frequency: int = 1
    print_every: int = 0


    # Training related
    optimizer_name: str = "AdamW"
    lr: float = 3e-4
    weight_decay: float = 0.01
    loss_fn_name: str = "CrossEntropy"
    batch_size: int = 512
    warmup_steps: int = 10
    scheduler_step: int = 200
    scheduler_gamma: float = .8
    early_stopping: bool = False
    early_stopping_patience: int = 2
    num_epochs: int = 50
    max_grad_norm = 1.0

    # Saving related
    parent_dir: str = ""
    save_dir: str = 'Checkpoints'
    save_checkpoints: bool = True
    save_cp_frequency: int = 5

    # Properties specific to video transformers
    is_video_transformer: bool = False
    video_tubelet_depth: int = 2
    video_num_frames: int = 32


    @classmethod
    def from_dict(cls, config_dict: Dict[str, Any]):
        return cls(**config_dict)


"""
Prisma Repo
By Sonia Joseph

Copyright (c) Sonia Joseph. All rights reserved.

Inspired by TransformerLens. Some functions have been adapted from the TransformerLens project.
For more information on TransformerLens, visit: https://github.com/neelnanda-io/TransformerLens
"""

import numpy as np
import torch
from typing import Dict, Optional, Tuple, Union
from jaxtyping import Float, Int
import torch.nn.functional as F




def gelu_new(
    input: Float[torch.Tensor, "batch pos d_mlp"]
) -> Float[torch.Tensor, "batch pos d_mlp"]:
    # Implementation of GeLU used by GPT2 - subtly different from PyTorch's
    return (
        0.5
        * input
        * (
            1.0
            + torch.tanh(
                np.sqrt(2.0 / np.pi) * (input + 0.044715 * torch.pow(input, 3.0))
            )
        )
    )


def gelu_fast(
    input: Float[torch.Tensor, "batch pos d_mlp"]
) -> Float[torch.Tensor, "batch pos d_mlp"]:
    return (
        0.5
        * input
        * (1.0 + torch.tanh(input * 0.7978845608 * (1.0 + 0.044715 * input * input)))
    )


def solu(
    input: Float[torch.Tensor, "batch pos d_mlp"]
) -> Float[torch.Tensor, "batch pos d_mlp"]:
    """
    SoLU activation function as described by
    https://transformer-circuits.pub/2022/solu/index.html.

    LayerNorm implemented by the MLP class.
    """
    return input * F.softmax(input, dim=-1)


from dataclasses import dataclass
from vit_prisma.models.layers.transformer_block import TransformerBlock
import torch.nn as nn

@dataclass
class ImageConfig:
    image_size: int = 28
    patch_size: int = 7
    n_channels: int = 1

@dataclass
class TransformerConfig:
    hidden_dim: int = 128
    num_heads: int = 4
    num_layers: int = 4
    block_fn = TransformerBlock
    mlp_dim: int = hidden_dim * 4  # Use a computed default
    activation_name: str = 'GELU'
    attention_only: bool = False
    attn_hidden_layer: bool = True

@dataclass
class LayerNormConfig:
    qknorm: bool = False
    layer_norm_eps: float = 0.0

@dataclass
class DropoutConfig:
    patch: float = 0.0
    position: float = 0.0
    attention: float = 0.0
    proj: float = 0.0
    mlp: float = 0.0

@dataclass
class InitializationConfig:
    weight_type: str = 'he'
    cls_std: float = 1e-6
    pos_std: float = 0.02

@dataclass
class TrainingConfig:
    loss_fn_name: str = "CrossEntropy"
    lr: float = 1e-4
    num_epochs: int = 5
    batch_size: int = 64
    warmup_steps: int = 10
    weight_decay: float = 0.0
    max_grad_norm = None
    device: str = 'cuda'
    seed: int = 0
    optimizer_name: str = "AdamW"
    scheduler_step: int = 1200
    scheduler_gamma: float = .8

@dataclass
class LoggingConfig:
    log_dir: str = 'logs'
    log_frequency: int = 10
    print_every: int = 0
    use_wandb: bool = True
    wandb_project_name = 'mnist_test'

@dataclass
class SavingConfig:
    parent_dir: str = "/network/scratch/s/sonia.joseph/vit_prisma"
    save_dir: str = 'checkpoints'
    save_checkpoints: bool = True
    save_cp_frequency: int = 10

class ClassificationConfig:
    num_classes: int = 10
    global_pool: bool = False

@dataclass
class GlobalConfig:
    image: ImageConfig = ImageConfig()
    transformer: TransformerConfig = TransformerConfig()
    layernorm: LayerNormConfig = LayerNormConfig()
    dropout: DropoutConfig = DropoutConfig()
    init: InitializationConfig = InitializationConfig()
    training: TrainingConfig = TrainingConfig()
    logging: LoggingConfig = LoggingConfig()
    saving: SavingConfig = SavingConfig()
    classification: ClassificationConfig = ClassificationConfig()




# import unittest
# from torchvision import datasets, transforms
# from torch.nn import CrossEntropyLoss
# from vit_prisma.models.base_vit import BaseViT  # Assuming you have a ViT model defined somewhere

# from vit_prisma.configs.InductionConfig import GlobalConfig
# from vit_prisma.training.trainer import train
# from vit_prisma.dataloaders.induction import InductionDataset

# class TestTrainingFunction(unittest.TestCase):

#         # Load MNIST dataset
#         def setUp(self):

#             self.train_dataset = InductionDataset('train')
#             self.val_dataset = InductionDataset('test')
    
#             self.config = GlobalConfig()

#             # Define a simple model
#             self.model = BaseViT(self.config)

#         def test_train_function(self):
#             trained_model = train(self.model, self.config, self.train_dataset, self.val_dataset)
#             self.assertIsInstance(trained_model, type(self.model))

# if __name__ == '__main__':
#     unittest.main()
